#!/usr/bin/env python3
"""
Phase variants in a VCF with the WhatsHap algorithm

Read a VCF and one or more files with phase information (BAM/CRAM or VCF phased
blocks) and phase the variants. The phased VCF is written to standard output.
"""
import logging
import sys
import platform

from argparse import SUPPRESS
from collections import defaultdict
from copy import deepcopy

from contextlib import ExitStack
from pathlib import Path
from typing import (
    Optional,
    List,
    TextIO,
    Union,
    Dict,
    Sequence,
    Mapping,
    Tuple,
    Set,
    MutableSequence,
    IO,
)

from whatshap.vcf import VcfReader, PhasedVcfWriter, VcfError, VariantTable
from whatshap import __version__
from whatshap.core import (
    ReadSet,
    Pedigree,
    PedigreeDPTable,
    NumericSampleIds,
    PhredGenotypeLikelihoods,
    HapChatCore,
)
from whatshap.readselect import readselection
from whatshap.graph import ComponentFinder
from whatshap.pedigree import (
    PedReader,
    mendelian_conflict,
    UniformRecombinationCostComputer,
    GeneticMapRecombinationCostComputer,
    find_recombination,
    ParseError,
    RecombinationCostComputer,
    Trio,
)
from whatshap.timer import StageTimer
from whatshap.utils import plural_s, warn_once
from whatshap.cli import CommandLineError, log_memory_usage, PhasedInputReader
from whatshap.merge import ReadMerger, DoNothingReadMerger, ReadMergerBase

__author__ = "Murray Patterson, Alexander Schönhuth, Tobias Marschall, Marcel Martin"

logger = logging.getLogger(__name__)


def find_components(
    phased_positions: Sequence[int],
    reads: ReadSet,
    master_block: Optional[Sequence[int]] = None,
    heterozygous_positions: Optional[Mapping[int, Set[int]]] = None,
) -> Mapping[int, int]:
    """
    Return a dict that maps each variant position to the component it is in.
    Variants are considered to be in the same component if a read exists that
    covers both. A component is identified by the position of its leftmost
    variant.
    master_block -- List of positions in a "master block", i.e. all blocks containing
                    any of these positions are merged into one block.
    heterozygous_positions -- A dictionary mapping numeric sample ids to sets of
                              positions. Component building is then restricted to variants
                              at these positions. If None, all variants are used.
    """
    logger.debug("Finding connected components ...")
    assert phased_positions == sorted(phased_positions)

    # Find connected components.
    # A component is identified by the position of its leftmost variant.
    component_finder = ComponentFinder(phased_positions)
    phased_positions_set = set(phased_positions)
    for read in reads:
        if heterozygous_positions is None:
            positions = [
                variant.position for variant in read if variant.position in phased_positions_set
            ]
        else:
            positions = [
                variant.position
                for variant in read
                if (variant.position in phased_positions_set)
                and (variant.position in heterozygous_positions[read.sample_id])
            ]
        for position in positions[1:]:
            component_finder.merge(positions[0], position)
    if master_block is not None:
        for position in master_block[1:]:
            component_finder.merge(master_block[0], position)
    components = {position: component_finder.find(position) for position in phased_positions_set}
    return components


def find_largest_component(components: Mapping[int, int]) -> Sequence[int]:
    """
    Determine the largest component and return a sorted list of positions
    contained in it.
    components -- dictionary mapping position to block_id as returned by find_components.
    """
    blocks = defaultdict(list)
    for position, block_id in components.items():
        blocks[block_id].append(position)
    largest: List[int] = []
    for block in blocks.values():
        if len(block) > len(largest):
            largest = block
    largest.sort()
    return largest


def best_case_blocks(reads: ReadSet) -> Tuple[int, int]:
    """
    Given a list of core reads, determine the number of phased blocks that
    would result if each variant were actually phased.

    Return the number of connected components and non-singleton components.
    """
    positions = set()
    for read in reads:
        for variant in read:
            positions.add(variant.position)
    component_finder = ComponentFinder(positions)
    for read in reads:
        read_positions = [variant.position for variant in read]
        for position in read_positions[1:]:
            component_finder.merge(read_positions[0], position)
    # A dict that maps each component to the number of variants it contains
    component_sizes: Dict[int, int] = defaultdict(int)
    for position in positions:
        component_sizes[component_finder.find(position)] += 1
    non_singletons = [component for component, size in component_sizes.items() if size > 1]
    return len(component_sizes), len(non_singletons)


def select_reads(
    readset: ReadSet, max_coverage: int, preferred_source_ids: Optional[Set[int]]
) -> ReadSet:
    logger.debug(
        "Reducing coverage to at most %dX by selecting most informative reads ...", max_coverage
    )
    selected_indices = readselection(readset, max_coverage, preferred_source_ids)
    selected_reads = readset.subset(selected_indices)
    logger.info(
        "Selected %d most phase-informative reads covering %d variants",
        len(selected_reads),
        len(selected_reads.get_positions()),
    )
    return selected_reads


class ReadList:
    """Write a list of reads that have been used for phasing to a file"""

    def __init__(self, path: str):
        self._path = path
        self._file: Optional[IO] = None

    def __enter__(self):
        self._file = open(self._path, "w")
        print(
            "#readname",
            "source_id",
            "sample",
            "phaseset",
            "haplotype",
            "covered_variants",
            "first_variant_pos",
            "last_variant_pos",
            sep="\t",
            file=self._file,
        )
        return self

    def __exit__(self, *args):
        self._file.close()
        self._file = None

    def write(
        self,
        readset: ReadSet,
        bipartition: Sequence[int],
        sample_components: Mapping[str, Sequence[int]],
        numeric_sample_ids: NumericSampleIds,
    ) -> None:
        """
        readset -- reads to be written
        bipartition -- bipartition of reads, i.e. sequence with one entry from {0,1} for each read in readset
        sample_components -- a dictionary that maps each sample to its connected components

                Each component in turn is a dict that maps each variant position to a
                component, where a component is identified by the position of its
                left-most variant

        numeric_sample_ids -- core.NumericSampleIds object mapping sample names to numeric ids as stored in each read
        """
        if self._file is None:
            raise ValueError("Needs to be used as context manager (e.g. in a with statement")
        assert len(readset) == len(bipartition)
        numeric_id_to_name = numeric_sample_ids.inverse_mapping()
        for read, haplotype in zip(readset, bipartition):
            sample = numeric_id_to_name[read.sample_id]
            components = sample_components[sample]
            phaseset = components[read[0].position] + 1
            print(
                read.name,
                read.source_id,
                sample,
                phaseset,
                haplotype,
                len(read),
                read[0].position + 1,
                read[-1].position + 1,
                sep="\t",
                file=self._file,
            )


def setup_pedigree(ped_path: str, samples: Sequence[str]) -> Tuple[Sequence[Trio], Set[str]]:
    """
    Read in PED file to set up list of relationships.

    Return a pair (trios, pedigree_samples), where trios is a list of Trio
    objects and pedigree_samples is the set of all samples that are mentioned
    in the PED file (as individual, mother or father).

    ped_path -- path to PED file
    samples -- samples to be phased
    """
    trios = []
    pedigree_samples = set()
    for trio in PedReader(ped_path):
        if trio.child is None or trio.mother is None or trio.father is None:
            warn_once(
                logger,
                "Relationship %s/%s/%s ignored because at least one of the individuals is unknown.",
                trio.child,
                trio.mother,
                trio.father,
            )
            continue
        # if at least one individual is not in samples, skip trio
        if (
            (trio.mother not in samples)
            or (trio.father not in samples)
            or (trio.child not in samples)
        ):
            # happens in case --ped and --samples are used
            warn_once(
                logger,
                "Relationship %s/%s/%s ignored because at least one of the "
                "individuals was not given by --samples.",
                trio.child,
                trio.mother,
                trio.father,
            )
            continue

        trios.append(trio)
        pedigree_samples.add(trio.child)
        pedigree_samples.add(trio.father)
        pedigree_samples.add(trio.mother)

    return trios, pedigree_samples


def run_whatshap(
    phase_input_files: Sequence[str],
    variant_file: str,
    reference: Union[None, bool, str] = False,
    output: TextIO = sys.stdout,
    samples: Optional[Sequence[str]] = None,
    chromosomes: Optional[List[str]] = None,
    ignore_read_groups: bool = False,
    indels: bool = True,
    mapping_quality: int = 20,
    read_merging: bool = False,
    read_merging_error_rate: float = 0.15,
    read_merging_max_error_rate: float = 0.25,
    read_merging_positive_threshold: int = 1000000,
    read_merging_negative_threshold: int = 1000,
    max_coverage: int = 15,
    distrust_genotypes: bool = False,
    include_homozygous: bool = False,
    ped: Optional[str] = None,
    recombrate: float = 1.26,
    genmap: Optional[str] = None,
    genetic_haplotyping: bool = True,
    recombination_list_filename: Optional[str] = None,
    tag: str = "PS",
    read_list_filename: Optional[str] = None,
    gl_regularizer: Optional[float] = None,
    gtchange_list_filename: Optional[str] = None,
    default_gq: int = 30,
    write_command_line_header: bool = True,
    use_ped_samples: bool = False,
    algorithm: str = "whatshap",
) -> None:
    """
    Run WhatsHap.

    phase_input_files -- list of paths to BAM/CRAM/VCF files
    variant_file -- path to input VCF
    reference -- path to reference FASTA. If False: skip realignment. If None: complain if reference needed.
    output -- path to output VCF or a file-like object
    samples -- names of samples to phase. an empty list means: phase all samples
    chromosomes -- names of chromosomes to phase. an empty list means: phase all chromosomes
    ignore_read_groups
    mapping_quality -- discard reads below this mapping quality
    read_merging -- whether or not to merge reads
    read_merging_error_rate -- probability that a nucleotide is wrong
    read_merging_max_error_rate -- max error rate on edge of merge graph considered
    read_merging_positive_threshold -- threshold on the ratio of the two probabilities
    read_merging_negative_threshold -- threshold on the opposite ratio of positive threshold
    max_coverage
    distrust_genotypes
    include_homozygous
    genetic_haplotyping -- in ped mode, merge disconnected blocks based on genotype status
    recombination_list_filename -- filename to write putative recombination events to
    tag -- How to store phasing info in the VCF, can be 'PS' or 'HP'
    read_list_filename -- name of file to write list of used reads to
    algorithm -- algorithm to use, can be 'whatshap' or 'hapchat'
    gl_regularizer -- float to be passed as regularization constant to GenotypeLikelihoods.as_phred
    gtchange_list_filename -- filename to write list of changed genotypes to
    default_gq -- genotype likelihood to be used when GL or PL not available
    write_command_line_header -- whether to add a ##commandline header to the output VCF
    """

    if algorithm == "hapchat" and ped is not None:
        raise CommandLineError("The hapchat algorithm cannot do pedigree phasing")
    if samples is None:
        samples = []

    timers = StageTimer()
    logger.info(f"This is WhatsHap {__version__} running under Python {platform.python_version()}")
    numeric_sample_ids = NumericSampleIds()
    command_line: Optional[str]
    if write_command_line_header:
        command_line = "(whatshap {}) {}".format(__version__, " ".join(sys.argv[1:]))
    else:
        command_line = None

    read_merger: ReadMergerBase
    if read_merging:
        read_merger = ReadMerger(
            read_merging_error_rate,
            read_merging_max_error_rate,
            read_merging_positive_threshold,
            read_merging_negative_threshold,
        )
    else:
        read_merger = DoNothingReadMerger()

    with ExitStack() as stack:
        logger.debug("Creating PhasedInputReader")
        phased_input_reader = stack.enter_context(
            PhasedInputReader(
                phase_input_files,
                None if reference is False else reference,
                numeric_sample_ids,
                ignore_read_groups,
                mapq_threshold=mapping_quality,
                indels=indels,
            )
        )
        show_phase_vcfs = phased_input_reader.has_vcfs

        if phased_input_reader.has_alignments and reference is None:
            raise CommandLineError(
                "A reference FASTA needs to be provided with -r/--reference; "
                "or use --no-reference at the expense of phasing quality."
            )

        logger.debug("Creating PhasedVcfWriter")
        try:
            vcf_writer = stack.enter_context(
                PhasedVcfWriter(
                    command_line=command_line,
                    in_path=variant_file,
                    out_file=output,
                    tag=tag,
                    indels=indels,
                )
            )
        except (OSError, VcfError) as e:
            raise CommandLineError(e)

        # Only read genotype likelihoods from VCFs when distrusting genotypes
        vcf_reader = stack.enter_context(
            VcfReader(variant_file, indels=indels, genotype_likelihoods=distrust_genotypes)
        )

        if ignore_read_groups and not samples and len(vcf_reader.samples) > 1:
            raise CommandLineError(
                "When using --ignore-read-groups on a VCF with "
                "multiple samples, --sample must also be used."
            )

        if not samples:
            samples = vcf_reader.samples

        # if --use-ped-samples is set, use only samples from PED file
        if ped is not None and use_ped_samples:
            samples = PedReader(ped).samples()

        assert samples is not None
        raise_if_any_sample_not_in_vcf(vcf_reader, samples)

        recombination_cost_computer = make_recombination_cost_computer(ped, genmap, recombrate)

        families, family_trios = setup_families(samples, ped, max_coverage)
        del samples
        for trios in family_trios.values():
            for trio in trios:
                # Ensure that all mentioned individuals have a numeric id
                if trio.child is not None:
                    _ = numeric_sample_ids[trio.child]

        read_list = None
        if read_list_filename:
            read_list = stack.enter_context(ReadList(read_list_filename))
            if algorithm == "hapchat":
                logger.warning(
                    "On which haplotype a read occurs in the inferred solution is not yet "
                    "implemented in hapchat, and so the corresponding column in the "
                    "read list file contains no information about this"
                )

        with timers("parse_phasing_vcfs"):
            # TODO should this be done in PhasedInputReader.__init__?
            phased_input_reader.read_vcfs()

        superreads: Dict[str, ReadSet]
        components: Dict
        for variant_table in timers.iterate("parse_vcf", vcf_reader):
            chromosome = variant_table.chromosome
            if chromosomes and chromosome not in chromosomes:
                logger.info(
                    "Leaving chromosome %r unchanged "
                    "(present in VCF but not requested by --chromosome)",
                    chromosome,
                )
                with timers("write_vcf"):
                    superreads, components = dict(), dict()
                    vcf_writer.write(chromosome, superreads, components)
                continue

            # These two variables hold the phasing results for all samples
            superreads, components = dict(), dict()

            # Iterate over all families to process, i.e. a separate DP table is created
            # for each family.
            # TODO: Can the body of this loop be factored out into a phase_family function?
            for representative_sample, family in sorted(families.items()):
                logger.info("")
                if len(family) == 1:
                    logger.info(
                        "# Working on contig %s in individual %s", chromosome, representative_sample
                    )
                else:
                    logger.info(
                        "# Working on contig %s in family individuals %s",
                        chromosome,
                        ",".join(family),
                    )
                max_coverage_per_sample = max(1, max_coverage // len(family))
                logger.debug("Using maximum coverage per sample of %dX", max_coverage_per_sample)
                trios = family_trios[representative_sample]
                assert len(family) == 1 or len(trios) > 0

                homozygous_positions, phasable_variant_table = find_phaseable_variants(
                    family, include_homozygous, trios, variant_table
                )

                # Get the reads belonging to each sample
                readsets = dict()  # TODO this could become a list
                for sample in family:
                    with timers("read_bam"):
                        readset, vcf_source_ids = phased_input_reader.read(
                            chromosome, phasable_variant_table.variants, sample
                        )

                    # TODO: Read selection done w.r.t. all variants, where using heterozygous
                    #  variants only would probably give better results.
                    with timers("select"):
                        readset = readset.subset(
                            [i for i, read in enumerate(readset) if len(read) >= 2]
                        )
                        logger.info(
                            "Kept %d reads that cover at least two variants each", len(readset)
                        )
                        merged_reads = read_merger.merge(readset)
                        selected_reads = select_reads(
                            merged_reads,
                            max_coverage_per_sample,
                            preferred_source_ids=vcf_source_ids,
                        )

                    readsets[sample] = selected_reads
                    if len(family) == 1 and not distrust_genotypes:
                        # When having a pedigree (len(family) > 1), blocks are also merged after
                        # phasing based on the pedigree information and these statistics are not
                        # so useful. When distrust_genotypes, genotypes can change during phasing
                        # and so can the block structure. So don't print these stats in those cases
                        log_best_case_phasing_info(readset, selected_reads)

                all_reads = merge_readsets(readsets)

                # Determine which variants can (in principle) be phased
                accessible_positions = sorted(all_reads.get_positions())
                logger.debug(
                    "Variants covered by at least one phase-informative "
                    "read in at least one individual after read selection: %d",
                    len(accessible_positions),
                )
                if len(family) > 1 and genetic_haplotyping:
                    # In case of genetic haplotyping, also retain all positions homozygous
                    # in at least one individual (because they might be phased based on genotypes)
                    accessible_positions = sorted(
                        set(accessible_positions).union(homozygous_positions)
                    )
                    logger.info(
                        "Variants either covered by phase-informative read or homozygous "
                        "in at least one individual: %d",
                        len(accessible_positions),
                    )

                # Keep only accessible positions
                phasable_variant_table.subset_rows_by_position(accessible_positions)
                assert len(phasable_variant_table.variants) == len(accessible_positions)

                pedigree = create_pedigree(
                    default_gq,
                    distrust_genotypes,
                    family,
                    gl_regularizer,
                    numeric_sample_ids,
                    phasable_variant_table,
                    trios,
                )
                recombination_costs = recombination_cost_computer.compute(accessible_positions)

                # Finally, run phasing algorithm
                with timers("phase"):
                    problem_name = "MEC" if len(family) == 1 else "PedMEC"
                    logger.info(
                        "Phasing %d sample%s by solving the %s problem ...",
                        len(family),
                        plural_s(len(family)),
                        problem_name,
                    )

                    dp_table: Union[HapChatCore, PedigreeDPTable]
                    if algorithm == "hapchat":
                        dp_table = HapChatCore(all_reads)
                    else:
                        dp_table = PedigreeDPTable(
                            all_reads,
                            recombination_costs,
                            pedigree,
                            distrust_genotypes,
                            accessible_positions,
                        )

                    superreads_list, transmission_vector = dp_table.get_super_reads()
                    logger.debug("%s cost: %d", problem_name, dp_table.get_optimal_cost())

                with timers("components"):
                    overall_components = compute_overall_components(
                        accessible_positions,
                        all_reads,
                        distrust_genotypes,
                        family,
                        genetic_haplotyping,
                        homozygous_positions,
                        numeric_sample_ids,
                        superreads_list,
                    )
                    log_component_stats(overall_components, len(accessible_positions))

                if recombination_list_filename:
                    n_recombinations = write_recombination_list(
                        recombination_list_filename,
                        chromosome,
                        accessible_positions,
                        overall_components,
                        recombination_costs,
                        transmission_vector,
                        trios,
                    )
                    logger.info("Total no. of detected recombination events: %d", n_recombinations)

                # Superreads in superreads_list are in the same order as individuals were added to the pedigree
                for sample, sample_superreads in zip(family, superreads_list):
                    superreads[sample] = sample_superreads
                    assert len(sample_superreads) == 2
                    assert (
                        sample_superreads[0].sample_id
                        == sample_superreads[1].sample_id
                        == numeric_sample_ids[sample]
                    )
                    # identical for all samples
                    components[sample] = overall_components

                if read_list:
                    read_list.write(
                        all_reads,
                        dp_table.get_optimal_partitioning(),
                        components,
                        numeric_sample_ids,
                    )

            with timers("write_vcf"):
                logger.debug("Writing phasing result to output VCF")
                changed_genotypes = vcf_writer.write(chromosome, superreads, components)
                if changed_genotypes:
                    assert distrust_genotypes
                    logger.info("Changed %d genotypes while writing VCF", len(changed_genotypes))

            if gtchange_list_filename:
                logger.info("Writing list of changed genotypes to %r", gtchange_list_filename)
                write_changed_genotypes(gtchange_list_filename, changed_genotypes)

            logger.debug("Chromosome %r finished", chromosome)

    log_time_and_memory_usage(timers, show_phase_vcfs=show_phase_vcfs)


def compute_overall_components(
    accessible_positions: Sequence[int],
    all_reads: ReadSet,
    distrust_genotypes: bool,
    family: Sequence[str],
    genetic_haplotyping: bool,
    homozygous_positions: Sequence[int],
    numeric_sample_ids: NumericSampleIds,
    superreads_list: Sequence[ReadSet],
) -> Mapping[int, int]:

    master_block = None
    heterozygous_positions_by_sample: Optional[Dict[int, Set[int]]] = None
    accessible_positions_set = set(accessible_positions)
    # If we distrusted genotypes, we need to re-determine which sites are homo-/heterozygous after phasing
    if distrust_genotypes:
        hom_in_any_sample = set()
        heterozygous_positions_by_sample = {}
        heterozygous_gts = frozenset({(0, 1), (1, 0)})
        homozygous_gts = frozenset({(0, 0), (1, 1)})
        for sample, sample_superreads in zip(family, superreads_list):
            hets = set()
            for v1, v2 in zip(*sample_superreads):
                assert v1.position == v2.position
                if v1.position not in accessible_positions_set:
                    continue
                gt = (v1.allele, v2.allele)
                if gt in heterozygous_gts:
                    hets.add(v1.position)
                elif gt in homozygous_gts:
                    hom_in_any_sample.add(v1.position)
            heterozygous_positions_by_sample[numeric_sample_ids[sample]] = hets
        if len(family) > 1 and genetic_haplotyping:
            master_block = sorted(hom_in_any_sample)
    else:
        if len(family) > 1 and genetic_haplotyping:
            master_block = sorted(set(homozygous_positions).intersection(accessible_positions_set))
    return find_components(
        accessible_positions, all_reads, master_block, heterozygous_positions_by_sample
    )


def log_component_stats(components: Mapping[int, int], n_accessible_positions: int) -> None:
    n_phased_blocks = len(set(components.values()))
    largest = find_largest_component(components)
    if largest:
        logger.info(
            "%s",
            f"Largest block contains {len(largest)} variants"
            f" ({len(largest) / n_accessible_positions:.1%} of accessible variants)"
            f" between position {largest[0] + 1} and {largest[-1] + 1}",
        )
    else:
        logger.info(f"No. of phased blocks: {n_phased_blocks}")


def log_best_case_phasing_info(readset: ReadSet, selected_reads: ReadSet) -> None:
    (n_best_case_blocks, n_best_case_nonsingleton_blocks) = best_case_blocks(readset)
    (n_best_case_blocks_cov, n_best_case_nonsingleton_blocks_cov) = best_case_blocks(selected_reads)
    logger.info(
        "Best-case phasing would result in %d non-singleton phased block%s (%d singletons). ",
        n_best_case_nonsingleton_blocks_cov,
        plural_s(n_best_case_nonsingleton_blocks_cov),
        n_best_case_blocks_cov - n_best_case_nonsingleton_blocks_cov,
    )
    logger.debug(
        "... would be %d non-singleton phased blocks without read selection",
        n_best_case_nonsingleton_blocks,
    )


def raise_if_any_sample_not_in_vcf(vcf_reader: VcfReader, samples: Sequence[str]) -> None:
    vcf_sample_set = set(vcf_reader.samples)
    for sample in samples:
        if sample not in vcf_sample_set:
            raise CommandLineError(f"Sample {sample!r} requested on command-line not found in VCF")


def setup_families(
    samples: Sequence[str], ped_path: Optional[str], max_coverage: int
) -> Tuple[Mapping[str, Sequence[str]], Mapping[str, Sequence[Trio]]]:
    """
    Return families, family_trios pair.

    families maps a family representative to a list of family members

    family_trios maps a family representative to a list of trios in this family
    """

    # list of all trios across all families

    # Keep track of connected components (aka families) in the pedigree
    family_finder = ComponentFinder(samples)
    if ped_path is not None:
        all_trios, pedigree_samples = setup_pedigree(ped_path, samples)
        for trio in all_trios:
            if trio.father is not None:
                family_finder.merge(trio.father, trio.child)
            if trio.mother is not None:
                family_finder.merge(trio.mother, trio.child)
    else:
        all_trios = []

    # map family representatives to lists of family members
    families: Mapping[str, MutableSequence[str]] = defaultdict(list)
    for sample in samples:
        families[family_finder.find(sample)].append(sample)

    # map family representatives to lists of trios for this family
    family_trios: Mapping[str, MutableSequence[Trio]] = defaultdict(list)
    for trio in all_trios:
        family_trios[family_finder.find(trio.child)].append(trio)
    logger.info(
        "Working on %d sample%s from %d famil%s",
        len(samples),
        plural_s(len(samples)),
        len(families),
        "y" if len(families) == 1 else "ies",
    )

    largest_trio_count = max([0] + [len(trio_list) for trio_list in family_trios.values()])
    if max_coverage + 2 * largest_trio_count > 23:
        logger.warning(
            "The maximum coverage is too high! "
            "WhatsHap may take a long time to finish and require a huge amount of memory."
        )
    return families, family_trios


def make_recombination_cost_computer(
    ped: Optional[str], genmap: Optional[str], recombrate: float
) -> RecombinationCostComputer:
    if ped and genmap:
        logger.info("Using region-specific recombination rates from genetic map %s.", genmap)
        try:
            return GeneticMapRecombinationCostComputer(genmap)
        except ParseError as e:
            raise CommandLineError(e)
    else:
        if ped:
            logger.info("Using uniform recombination rate of %g cM/Mb.", recombrate)
        return UniformRecombinationCostComputer(recombrate)


def find_phaseable_variants(
    family: Sequence[str],
    include_homozygous: bool,
    trios: Sequence[Trio],
    variant_table: VariantTable,
) -> Tuple[Sequence[int], VariantTable]:

    # variant indices with at least one missing genotype
    missing_genotypes = set()
    # variant indices with at least one heterozygous genotype
    heterozygous = set()
    # variant indices with at least one homozygous genotype
    homozygous = set()
    # determine which variants have missing/heterozygous/homozygous genotypes in any sample
    for sample in family:
        genotypes = variant_table.genotypes_of(sample)
        for index, gt in enumerate(genotypes):
            if gt.is_none():
                missing_genotypes.add(index)
            elif not gt.is_homozygous():
                heterozygous.add(index)
            else:
                assert gt.is_diploid_and_biallelic()
                homozygous.add(index)
    # determine which variants have Mendelian conflicts
    # variant indices with at least one Mendelian conflict
    mendelian_conflicts = find_mendelian_conflicts(trios, variant_table)
    # retain variants that are heterozygous in at least one individual (anywhere in the pedigree)
    # and do not have neither missing genotypes nor Mendelian conflicts
    if include_homozygous:
        to_retain = set(range(len(variant_table)))
    else:
        to_retain = heterozygous
    to_retain = to_retain.difference(missing_genotypes).difference(mendelian_conflicts)
    # discard every variant that is not to be retained
    to_discard = set(range(len(variant_table))).difference(to_retain)
    # Determine positions of selected variants that are homozygous in at least one individual.
    # These are used later to merge blocks containing these variants into one block (since
    # they are connected by "genetic haplotyping").
    homozygous_positions = [
        variant_table.variants[i].position for i in to_retain.intersection(homozygous)
    ]
    phasable_variant_table = deepcopy(variant_table)
    # Remove calls to be discarded from variant table
    phasable_variant_table.remove_rows_by_index(to_discard)

    if len(family) == 1:
        logger.info(
            "Found %d usable%s variants (%d skipped due to missing genotypes)",
            len(phasable_variant_table),
            "" if include_homozygous else " heterozygous",
            len(missing_genotypes),
        )
    else:
        logger.info(
            "Found %d usable variants (%d skipped due to Mendelian conflicts)",
            len(phasable_variant_table),
            len(mendelian_conflicts),
        )
    return homozygous_positions, phasable_variant_table


def log_time_and_memory_usage(timers, show_phase_vcfs):
    total_time = timers.total()
    logger.info("\n# Resource usage")
    log_memory_usage()
    # fmt: off
    logger.info("Time spent reading BAM/CRAM:                 %6.1f s", timers.elapsed("read_bam"))
    logger.info("Time spent parsing VCF:                      %6.1f s", timers.elapsed("parse_vcf"))
    if show_phase_vcfs:
        logger.info("Time spent parsing input phasings from VCFs: %6.1f s", timers.elapsed("parse_phasing_vcfs"))
    logger.info("Time spent selecting reads:                  %6.1f s", timers.elapsed("select"))
    logger.info("Time spent phasing:                          %6.1f s", timers.elapsed("phase"))
    logger.info("Time spent writing VCF:                      %6.1f s", timers.elapsed("write_vcf"))
    logger.info("Time spent finding components:               %6.1f s", timers.elapsed("components"))
    logger.info("Time spent on rest:                          %6.1f s", total_time - timers.sum())
    logger.info("Total elapsed time:                          %6.1f s", total_time)
    # fmt: on


def merge_readsets(readsets) -> ReadSet:
    all_reads = ReadSet()
    for sample, readset in readsets.items():
        for read in readset:
            assert read.is_sorted(), "Add a read.sort() here"
            all_reads.add(read)
    all_reads.sort()
    return all_reads


def create_pedigree(
    default_gq,
    distrust_genotypes,
    family,
    gl_regularizer,
    numeric_sample_ids,
    phasable_variant_table,
    trios,
):
    pedigree = Pedigree(numeric_sample_ids)
    for sample in family:
        # If distrusting genotypes, we pass genotype likelihoods on to pedigree object
        if distrust_genotypes:
            genotype_likelihoods = []
            for gt, gl in zip(
                phasable_variant_table.genotypes_of(sample),
                phasable_variant_table.genotype_likelihoods_of(sample),
            ):
                assert gt.is_diploid_and_biallelic()
                if gl is None:
                    # all genotypes get default_gq as genotype likelihood, exept the called genotype ...
                    x = [default_gq] * 3
                    # ... which gets a 0
                    x[gt.get_index()] = 0
                    genotype_likelihoods.append(PhredGenotypeLikelihoods(x))
                else:
                    genotype_likelihoods.append(gl.as_phred(regularizer=gl_regularizer))
        else:
            genotype_likelihoods = None
        pedigree.add_individual(
            sample, phasable_variant_table.genotypes_of(sample), genotype_likelihoods
        )
    for trio in trios:
        pedigree.add_relationship(father_id=trio.father, mother_id=trio.mother, child_id=trio.child)
    return pedigree


def find_mendelian_conflicts(trios: Sequence[Trio], variant_table: VariantTable) -> Set[int]:
    mendelian_conflicts = set()
    for trio in trios:
        if trio.mother is None or trio.father is None:
            continue
        genotypes_mother = variant_table.genotypes_of(trio.mother)
        genotypes_father = variant_table.genotypes_of(trio.father)
        genotypes_child = variant_table.genotypes_of(trio.child)

        for index, (gt_mother, gt_father, gt_child) in enumerate(
            zip(genotypes_mother, genotypes_father, genotypes_child)
        ):
            if (not gt_mother.is_none()) and (not gt_father.is_none()) and (not gt_child.is_none()):
                if mendelian_conflict(gt_mother, gt_father, gt_child):
                    mendelian_conflicts.add(index)
    return mendelian_conflicts


def write_changed_genotypes(gtchange_list_filename, changed_genotypes):
    with open(gtchange_list_filename, "w") as f:
        print(
            "#sample", "chromosome", "position", "REF", "ALT", "old_gt", "new_gt", sep="\t", file=f
        )
        for changed_genotype in changed_genotypes:
            print(
                changed_genotype.sample,
                changed_genotype.chromosome,
                changed_genotype.variant.position,
                changed_genotype.variant.reference_allele,
                changed_genotype.variant.alternative_allele,
                repr(changed_genotype.old_gt),
                repr(changed_genotype.new_gt),
                sep="\t",
                file=f,
            )


def write_recombination_list(
    path: Union[str, Path],
    chromosome: str,
    accessible_positions: Sequence[int],
    overall_components: Mapping[int, int],
    recombination_costs: Sequence[int],
    transmission_vector: Sequence[int],
    trios: Sequence[Trio],
) -> int:
    """Return total number of recombinations"""

    transmission_vector_trio: Mapping[str, MutableSequence[int]] = defaultdict(list)
    for transmission_vector_value in transmission_vector:
        for trio in trios:
            value = transmission_vector_value % 4
            transmission_vector_value = transmission_vector_value // 4
            transmission_vector_trio[trio.child].append(value)
    with open(path, "w") as f:
        n = 0
        print(
            "#child_id",
            "chromosome",
            "position1",
            "position2",
            "transmitted_hap_father1",
            "transmitted_hap_father2",
            "transmitted_hap_mother1",
            "transmitted_hap_mother2",
            "recombination_cost",
            file=f,
        )
        for trio in trios:
            recombination_events = find_recombination(
                transmission_vector_trio[trio.child],
                overall_components,
                accessible_positions,
                recombination_costs,
            )
            for e in recombination_events:
                print(
                    trio.child,
                    chromosome,
                    e.position1 + 1,
                    e.position2 + 1,
                    e.transmitted_hap_father1,
                    e.transmitted_hap_father2,
                    e.transmitted_hap_mother1,
                    e.transmitted_hap_mother2,
                    e.recombination_cost,
                    file=f,
                )
            n += len(recombination_events)
    return n


# fmt: off
def add_arguments(parser):
    arg = parser.add_argument
    # Positional arguments
    arg("variant_file", metavar="VCF",
        help="VCF or BCF file with variants to be phased (can be gzip-compressed)")
    arg("phase_input_files", nargs="*", metavar="PHASEINPUT",
        help="BAM, CRAM, VCF or BCF file(s) with phase information, either through "
        "sequencing reads (BAM, CRAM) or through phased blocks (VCF, BCF)")

    arg("-o", "--output", default=sys.stdout,
        help="Output VCF file. Add .gz to the file name to get compressed output. "
        "If omitted, use standard output.")
    arg("--reference", "-r", metavar="FASTA",
        help="Reference file. Must be accompanied by .fai index (create with samtools faidx)")
    arg("--no-reference", action="store_true", default=False,
        help="Detect alleles without requiring a reference, at the expense of phasing quality "
        "(in particular for long reads)")
    arg("--tag", choices=("PS", "HP"), default="PS",
        help="Store phasing information with PS tag (standardized) or "
        "HP tag (used by GATK ReadBackedPhasing) (default: %(default)s)")
    arg("--output-read-list", metavar="FILE", default=None, dest="read_list_filename",
        help="Write reads that have been used for phasing to FILE.")
    arg("--algorithm", choices=("whatshap", "hapchat"), default="whatshap",
        help="Phasing algorithm to use (default: %(default)s)")

    arg = parser.add_argument_group("Input pre-processing, selection and filtering").add_argument
    arg("--merge-reads", dest="read_merging", default=False, action="store_true",
        help="Merge reads which are likely to come from the same haplotype "
        "(default: do not merge reads)")
    arg("--max-coverage", "-H", metavar="MAXCOV", type=int,
        dest="max_coverage_was_used", help=SUPPRESS)
    arg("--internal-downsampling", metavar="COVERAGE", dest="max_coverage", default=15, type=int,
        help="Coverage reduction parameter in the internal core phasing algorithm. "
        "Higher values increase runtime *exponentially* while possibly improving phasing "
        "quality marginally. Avoid using this in the normal case! (default: %(default)s)")
    arg("--mapping-quality", "--mapq", metavar="QUAL",
        default=20, type=int, help="Minimum mapping quality (default: %(default)s)")
    arg("--indels", dest="indels", default=False, action="store_true",
        help="Also phase indels (default: do not phase indels)")
    arg("--ignore-read-groups", default=False, action="store_true",
        help="Ignore read groups in BAM/CRAM header and assume all reads come "
        "from the same sample.")
    arg("--sample", dest="samples", metavar="SAMPLE", default=[], action="append",
        help="Name of a sample to phase. If not given, all samples in the "
        "input VCF are phased. Can be used multiple times.")
    arg("--chromosome", dest="chromosomes", metavar="CHROMOSOME", default=[], action="append",
        help="Name of chromosome to phase. If not given, all chromosomes in the "
        "input VCF are phased. Can be used multiple times.")

    arg = parser.add_argument_group(
        "Read merging",
        "The options in this section are only active when --merge-reads is used"
    ).add_argument
    arg("--error-rate", dest="read_merging_error_rate",
        type=float, default=0.15,
        help="The probability that a nucleotide is wrong in read merging model "
        "(default: %(default)s).")
    arg("--maximum-error-rate", dest="read_merging_max_error_rate",
        type=float, default=0.25,
        help="The maximum error rate of any edge of the read merging graph "
        "before discarding it (default: %(default)s).")
    arg("--threshold", dest="read_merging_positive_threshold",
        type=int, default=1000000,
        help="The threshold of the ratio between the probabilities that a pair "
        "of reads come from the same haplotype and different haplotypes in the "
        "read merging model (default: %(default)s).")
    arg("--negative-threshold", dest="read_merging_negative_threshold",
        type=int, default=1000,
        help="The threshold of the ratio between the probabilities that a pair "
        "of reads come from different haplotypes and the same haplotype in the "
        "read merging model (default: %(default)s).")

    arg = parser.add_argument_group(
        "Genotyping",
        "These options are only used when --distrust-genotypes is used"
    ).add_argument
    arg("--full-genotyping", action="store_true", default=False, help=SUPPRESS)
    arg("--distrust-genotypes", dest="distrust_genotypes",
        action="store_true", default=False,
        help="Allow switching variants from hetero- to homozygous in an "
        "optimal solution (see documentation).")
    arg("--include-homozygous", dest="include_homozygous",
        action="store_true", default=False,
        help="Also work on homozygous variants, which might be turned to "
        "heterozygous")
    arg("--default-gq", type=int, default=30,
        help="Default genotype quality used as cost of changing a genotype "
        "when no genotype likelihoods are available (default %(default)s)")
    arg("--gl-regularizer", type=float, default=None,
        help="Constant (float) to be used to regularize genotype likelihoods read "
        "from input VCF (default %(default)s).")
    arg("--changed-genotype-list", metavar="FILE", dest="gtchange_list_filename", default=None,
        help="Write list of changed genotypes to FILE.")

    arg = parser.add_argument_group("Pedigree phasing").add_argument
    arg("--ped", metavar="PED/FAM",
        help="Use pedigree information in PED file to improve phasing "
        "(switches to PedMEC algorithm). Columns 2, 3, 4 must refer to child, "
        "mother, and father sample names as used in the VCF and BAM/CRAM. "
        "Other columns are ignored.")
    arg("--recombination-list", metavar="FILE", dest="recombination_list_filename", default=None,
        help="Write putative recombination events to FILE.")
    arg("--recombrate", metavar="RECOMBRATE", type=float, default=1.26,
        help="Recombination rate in cM/Mb (used with --ped). If given, a constant recombination "
        "rate is assumed (default: %(default)gcM/Mb).")
    arg("--genmap", metavar="FILE",
        help="File with genetic map (used with --ped) to be used instead of constant recombination "
        "rate, i.e. overrides option --recombrate.")
    arg("--no-genetic-haplotyping", dest="genetic_haplotyping",
        action="store_false", default=True,
        help="Do not merge blocks that are not connected by reads (i.e. solely based on genotype "
        "status). Default: when in --ped mode, merge all blocks that contain at least one "
        "homozygous genotype in at least one individual into one block.")
    arg("--use-ped-samples", dest="use_ped_samples",
        action="store_true", default=False,
        help="Only work on samples mentioned in the provided PED file.")
# fmt: on


def validate(args, parser):
    if args.reference is not None and args.no_reference:
        parser.error("Options --reference and --no-reference cannot be used together")
    if args.ignore_read_groups and args.ped:
        parser.error("Option --ignore-read-groups cannot be used together with --ped")
    if args.genmap and not args.ped:
        parser.error("Option --genmap can only be used together with --ped")
    if args.genmap and (len(args.chromosomes) != 1):
        parser.error(
            "Option --genmap can only be used when working on exactly one "
            "chromosome (use --chromosome)"
        )
    if args.include_homozygous and not args.distrust_genotypes:
        parser.error("Option --include-homozygous can only be used with --distrust-genotypes.")
    if args.use_ped_samples and not args.ped:
        parser.error("Option --use-ped-samples can only be used when PED file is provided (--ped).")
    if args.use_ped_samples and args.samples:
        parser.error("Option --use-ped-samples cannot be used together with --samples")
    if len(args.phase_input_files) == 0 and not args.ped:
        parser.error("Not providing any PHASEINPUT files only allowed in --ped mode.")
    if args.max_coverage > 23:
        parser.error("Coverage downsampling parameter must not exceed 23.")
    if args.max_coverage_was_used is not None:
        logger.warning(
            "The --max-coverage and -H options are no longer supported. "
            "The coverage reduction parameter in the internal core phasing algorithm can now "
            "be adjusted with --internal-downsampling. Higher values increase runtime "
            "*exponentially* while possibly improving phasing quality marginally. "
            "Avoid using this in the normal case!"
        )
    if args.full_genotyping:
        parser.error(
            "The experimental --full-genotyping option has been removed. Instead, please run "
            "'whatshap genotype' prior to running 'whatshap phase'"
        )


def main(args):
    if args.no_reference:
        args.reference = False
    del args.no_reference
    del args.max_coverage_was_used
    del args.full_genotyping
    run_whatshap(**vars(args))



import io
from pytest import raises
from whatshap.pedigree import PedReader, Trio, ParseError


class TestPedReader:
    def test_parse(self):
        trios = list(PedReader("tests/data/pedigree.ped"))
        assert trios[0] == Trio(child="child1", mother="mother", father="father")
        assert trios[1] == Trio(child="child2", mother="mother", father="father")
        assert trios[2] == Trio(child="father", mother=None, father=None)
        assert trios[3] == Trio(child="mother", mother=None, father=None)
        assert trios[4] == Trio(child="orphan", mother=None, father=None)

    def test_parse_error(self):
        f = io.StringIO("buggy file")
        with raises(ParseError):
            list(PedReader(f))

    def test_duplicate_individual(self):
        f = io.StringIO("f1 c m f 0 1\nf1 c m f 0 1")
        with raises(ParseError):
            list(PedReader(f))



from pytest import fixture
from whatshap.core import (
    ReadSet,
    PedigreeDPTable,
    Pedigree,
    NumericSampleIds,
    PhredGenotypeLikelihoods,
    HapChatCore,
)
from whatshap.testhelpers import (
    string_to_readset,
    brute_force_phase,
    canonic_index_to_biallelic_gt,
    canonic_index_list_to_biallelic_gt_list,
)


@fixture(params=["whatshap", "hapchat"])
def algorithm(request):
    return request.param


def test_phase_empty_readset(algorithm):
    rs = ReadSet()
    recombcost = [1, 1]
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1])
    pedigree = Pedigree(NumericSampleIds())
    genotype_likelihoods = [None, None]
    pedigree.add_individual("individual0", genotypes, genotype_likelihoods)

    if algorithm == "hapchat":
        dp_table = HapChatCore(rs)
    else:
        dp_table = PedigreeDPTable(rs, recombcost, pedigree)

    _ = dp_table.get_super_reads()


def compare_phasing_brute_force(
    superreads, cost, partition, readset, all_heterozygous, weights=None, algorithm="whatshap"
):
    """Compares DPTable based phasing to brute force phasing and returns string representation of superreads."""
    assert len(superreads) == 2
    assert len(superreads[0]) == len(superreads[1])
    for v1, v2 in zip(*superreads):
        assert v1.position == v2.position
    haplotypes = tuple(sorted("".join(str(v.allele) for v in sr) for sr in superreads))
    (
        expected_cost,
        expected_partition,
        solution_count,
        expected_haplotype1,
        expected_haplotype2,
    ) = brute_force_phase(readset, all_heterozygous)
    print()
    print(superreads[0])
    print(superreads[1])
    print("Partition:", partition)
    print("Expected: ", expected_partition)
    print("Haplotypes:")
    print(haplotypes[0])
    print(haplotypes[1])
    print("Expected Haplotypes:")
    print(expected_haplotype1)
    print(expected_haplotype2)
    print("Cost:", cost)
    print("Expected cost:", expected_cost)
    # TODO: implement the reporting of an optimal partitioning in hapchat
    if algorithm == "whatshap":
        inverse_partition = [1 - p for p in partition]
        assert (partition == expected_partition) or (inverse_partition == expected_partition)
    assert solution_count == 1
    assert cost == expected_cost
    assert (haplotypes == (expected_haplotype1, expected_haplotype2)) or (
        haplotypes == (expected_haplotype2, expected_haplotype1)
    )


def check_phasing_single_individual(reads, algorithm="whatshap", weights=None):
    # 0) set up read set
    readset = string_to_readset(reads, weights)
    positions = readset.get_positions()

    # for hapchat
    if algorithm == "hapchat":
        dp_table = HapChatCore(readset)
        superreads = dp_table.get_super_reads()
        cost = dp_table.get_optimal_cost()
        partition = dp_table.get_optimal_partitioning()
        compare_phasing_brute_force(
            superreads[0][0], cost, partition, readset, True, weights, algorithm
        )
        return

    # 1) Phase using PedMEC code for single individual
    for all_heterozygous in [False, True]:
        recombcost = [1] * len(positions)  # recombination costs 1, should not occur
        pedigree = Pedigree(NumericSampleIds())
        genotype_likelihoods = [
            None if all_heterozygous else PhredGenotypeLikelihoods([0, 0, 0])
        ] * len(positions)
        pedigree.add_individual(
            "individual0",
            [canonic_index_to_biallelic_gt(1) for i in range(len(positions))],
            genotype_likelihoods,
        )  # all genotypes heterozygous
        dp_table = PedigreeDPTable(
            readset, recombcost, pedigree, distrust_genotypes=not all_heterozygous
        )
        superreads, transmission_vector = dp_table.get_super_reads()
        cost = dp_table.get_optimal_cost()
        # TODO: transmission vectors not returned properly, see issue 73
        assert len(set(transmission_vector)) == 1
        partition = dp_table.get_optimal_partitioning()
        compare_phasing_brute_force(
            superreads[0], cost, partition, readset, all_heterozygous, weights
        )

    # 2) Phase using PedMEC code for trios with two "empty" individuals (i.e. having no reads)
    for all_heterozygous in [False, True]:
        recombcost = [1] * len(positions)  # recombination costs 1, should not occur
        pedigree = Pedigree(NumericSampleIds())
        genotype_likelihoods = [
            None if all_heterozygous else PhredGenotypeLikelihoods([0, 0, 0])
        ] * len(positions)
        pedigree.add_individual(
            "individual0",
            [canonic_index_to_biallelic_gt(1) for _ in range(len(positions))],
            genotype_likelihoods,
        )  # all genotypes heterozygous
        pedigree.add_individual(
            "individual1",
            [canonic_index_to_biallelic_gt(1) for _ in range(len(positions))],
            genotype_likelihoods,
        )  # all genotypes heterozygous
        pedigree.add_individual(
            "individual2",
            [canonic_index_to_biallelic_gt(1) for _ in range(len(positions))],
            genotype_likelihoods,
        )  # all genotypes heterozygous
        pedigree.add_relationship("individual0", "individual1", "individual2")
        dp_table = PedigreeDPTable(
            readset, recombcost, pedigree, distrust_genotypes=not all_heterozygous
        )
        cost = dp_table.get_optimal_cost()
        superreads, transmission_vector = dp_table.get_super_reads()
        assert len(set(transmission_vector)) == 1
        partition = dp_table.get_optimal_partitioning()
        compare_phasing_brute_force(
            superreads[0], cost, partition, readset, all_heterozygous, weights
        )


def test_phase_trivial(algorithm):
    reads = """
          11
           01
        """
    check_phasing_single_individual(reads, algorithm)


def test_phase1(algorithm):
    reads = """
     10
     010
     010
    """
    check_phasing_single_individual(reads, algorithm)


def test_phase2(algorithm):
    reads = """
      1  11010
      00 00101
      001 0101
    """
    check_phasing_single_individual(reads, algorithm)


def test_phase3(algorithm):
    reads = """
      1  11010
      00 00101
      001 01010
    """
    check_phasing_single_individual(reads, algorithm)


def test_phase4(algorithm):
    reads = """
      1  11010
      00 00101
      001 01110
       1    111
    """
    check_phasing_single_individual(reads, algorithm)


def test_phase4a(algorithm):
    reads = """
      1  11010
      00 00101
      001 01110
       1    111
    """
    check_phasing_single_individual(reads, algorithm)


# note: these final two tests do not apply to hapchat because their
# (brute force solutions) are weighted phasings -- hapchat does not do
# weights (for the time being)
def test_phase5():
    reads = """
      0             0
      110111111111
      00100
           0001000000
           000
            10100
                  101
    """
    check_phasing_single_individual(reads)


def test_weighted_phasing1():
    reads = """
      1  11010
      00 00101
      001 01110
       1    111
    """
    weights = """
      2  13112
      11 23359
      223 56789
       2    111
    """
    check_phasing_single_individual(reads, "whatshap", weights)



from collections import defaultdict
import shutil
import pysam
import pytest

from whatshap.cli.haplotag import run_haplotag
from whatshap.cli import CommandLineError


def test_haplotag(tmp_path):
    outbam1 = tmp_path / "output1.bam"
    outbam2 = tmp_path / "output2.bam"
    outlist1 = tmp_path / "list1.tsv"
    outlist2 = tmp_path / "list2.tsv"

    # run haplotag with two vcfs containing opposite phasings (i.e. 1|0 - 0|1 ..)
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag.bam",
        haplotag_list=outlist1,
        output=outbam1,
    )
    run_haplotag(
        variant_file="tests/data/haplotag_2.vcf.gz",
        alignment_file="tests/data/haplotag.bam",
        haplotag_list=outlist2,
        output=outbam2,
    )
    for a1, a2 in zip(pysam.AlignmentFile(outbam1), pysam.AlignmentFile(outbam2)):
        assert a1.query_name == a2.query_name
        if a1.has_tag("HP"):
            assert a2.has_tag("HP")
            assert a1.get_tag("HP") != a2.get_tag("HP")
    for n, (line1, line2) in enumerate(zip(open(outlist1), open(outlist2))):
        fields1 = line1.split(sep="\t")
        fields2 = line2.split(sep="\t")
        assert len(fields1) == len(fields2) == 4
        if n == 0:
            continue
        queryname1, haplotype1, phaseset1, chromosome1 = fields1
        queryname2, haplotype2, phaseset2, chromosome2 = fields2
        assert queryname1 == queryname2
        assert (haplotype1 == haplotype2 == "none") or (haplotype1 != haplotype2)
        assert chromosome1 == chromosome2
    assert n == 20


@pytest.mark.parametrize(
    "vcf_path",
    [
        "tests/data/haplotag_2.vcf.gz",
        "tests/data/haplotag_with_csi_index.vcf.gz",
        "tests/data/haplotag_2.bcf",
    ],
)
def test_haplotag2(tmp_path, vcf_path):
    outbam = tmp_path / "output.bam"
    run_haplotag(variant_file=vcf_path, alignment_file="tests/data/haplotag.bam", output=outbam)
    ps_count = 0
    for alignment in pysam.AlignmentFile(outbam):
        if alignment.has_tag("PS"):
            ps_count += 1
        if alignment.has_tag("HP"):
            # simulated bam, we know from which haplotype each read originated (given in read name)
            true_ht = int(alignment.query_name[-1])
            assert true_ht == alignment.get_tag("HP")
    assert ps_count > 0


def test_haplotag_fails_if_index_missing(tmp_path):
    outbam = tmp_path / "output.bam"
    vcf_path = tmp_path / "vcf_without_index.vcf.gz"
    shutil.copy("tests/data/haplotag_1.vcf.gz", vcf_path)
    with pytest.raises(CommandLineError):
        run_haplotag(variant_file=vcf_path, alignment_file="tests/data/haplotag.bam", output=outbam)


def test_haplotag_cli_parser(tmp_path):
    """
    This test captures an error in the parser of the cli haplotag
    module - a wrong default value of "[]" instead of "None" for the
    "--regions" option leads to an empty output
    :return:
    """
    from whatshap.cli.haplotag import add_arguments as haplotag_add_arguments
    import argparse as argp

    outbam = tmp_path / "output.bam"
    parser = argp.ArgumentParser(description="haplotag_test_parser", prog="whatshap_pytest")
    haplotag_add_arguments(parser)
    haplotag_args = parser.parse_args(
        [
            "--no-reference",
            "--output",
            str(outbam),
            "tests/data/haplotag_2.vcf.gz",
            "tests/data/haplotag.bam",
        ]
    )
    haplotag_args.reference = False
    del haplotag_args.no_reference
    run_haplotag(**vars(haplotag_args))
    ps_count = 0
    for alignment in pysam.AlignmentFile(outbam):
        if alignment.has_tag("PS"):
            ps_count += 1
        if alignment.has_tag("HP"):
            # simulated BAM, we know from which haplotype each read originated (given in read name)
            true_ht = int(alignment.query_name[-1])
            assert true_ht == alignment.get_tag("HP")
    assert ps_count > 0


def test_haplotag_missing_SM_tag(tmp_path):
    outbam1 = tmp_path / "output1.bam"
    outbam2 = tmp_path / "output2.bam"
    outlist1 = tmp_path / "list1.tsv"
    outlist2 = tmp_path / "list2.tsv"

    # run haplotag
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag.bam",
        haplotag_list=outlist1,
        output=outbam1,
        ignore_read_groups=True,
    )
    # use copy of 'haplotag.bam' which lacks the 'SM' tag
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag_noSM.bam",
        haplotag_list=outlist2,
        output=outbam2,
        ignore_read_groups=True,
    )

    # results should be identical
    for a1, a2 in zip(pysam.AlignmentFile(outbam1), pysam.AlignmentFile(outbam2)):
        assert a1.query_name == a2.query_name
        if a1.has_tag("HP"):
            assert a2.has_tag("HP")
            assert a1.get_tag("HP") == a2.get_tag("HP")
        for n, (line1, line2) in enumerate(zip(open(outlist1), open(outlist2))):
            fields1 = line1.split(sep="\t")
            fields2 = line2.split(sep="\t")
            assert len(fields1) == len(fields2) == 4
            if n == 0:
                continue
            queryname1, haplotype1, phaseset1, chromosome1 = fields1
            queryname2, haplotype2, phaseset2, chromosome2 = fields2
            assert queryname1 == queryname2
            assert haplotype1 == haplotype2
            assert chromosome1 == chromosome2
        assert n == 20


def test_haplotag_missing_chromosome(tmp_path):
    outbam = tmp_path / "output.bam"

    # input BAM contains a chromosome for which there is no variant in the input VCF
    run_haplotag(
        variant_file="tests/data/haplotag.missing_chr.vcf.gz",
        alignment_file="tests/data/haplotag.large.bam",
        output=outbam,
    )
    ps_count = 0
    for alignment in pysam.AlignmentFile(outbam):
        if alignment.has_tag("PS"):
            ps_count += 1
    assert ps_count > 0


def test_contig_exists_in_bam_but_not_in_vcf_header(tmp_path):
    outbam = tmp_path / "output.bam"

    with pytest.raises(CommandLineError) as e:
        run_haplotag(
            variant_file="tests/data/haplotag.without_chr2.vcf.gz",
            alignment_file="tests/data/haplotag.large.bam",  # has reads mapped to chr2
            output=outbam,
        )
    assert "contig does not exist" in e.value.args[0]

    run_haplotag(
        variant_file="tests/data/haplotag.without_chr2.vcf.gz",
        alignment_file="tests/data/haplotag.large.bam",  # has reads mapped to chr2
        output=outbam,
        skip_missing_contigs=True,
    )


def test_haplotag_no_readgroups1(tmp_path):
    outbam1 = tmp_path / "output1.bam"
    outbam2 = tmp_path / "output2.bam"

    # run haplotag with/without --ignore-read-groups, results should be identical since files contain only data for one sample
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag.bam",
        output=outbam1,
    )
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag_noRG.bam",
        output=outbam2,
        ignore_read_groups=True,
    )
    for a1, a2 in zip(pysam.AlignmentFile(outbam1), pysam.AlignmentFile(outbam2)):
        assert a1.query_name == a2.query_name
        if a1.has_tag("HP"):
            assert a2.has_tag("HP")
            assert a1.get_tag("HP") == a2.get_tag("HP")


def test_haplotag_no_readgroups2():
    with pytest.raises((CommandLineError, ValueError)):
        # VCF contains multiple samples, there should be an error
        run_haplotag(
            variant_file="tests/data/haplotag_noRG.vcf.gz",
            alignment_file="tests/data/haplotag_noRG.bam",
            output="/dev/null",
            ignore_read_groups=True,
        )


def test_haplotag_sample_given(tmp_path):
    outbam = tmp_path / "output.bam"
    run_haplotag(
        variant_file="tests/data/haplotag_sample.vcf.gz",
        alignment_file="tests/data/haplotag_sample.bam",
        given_samples=["mother"],
        output=outbam,
    )
    for alignment in pysam.AlignmentFile(outbam):
        if alignment.get_tag("RG") == "mother":
            assert alignment.has_tag("HP")
        else:
            assert not alignment.has_tag("HP")


def haplotag_different_sorting(tmp_path):
    outbam1 = tmp_path / "output1.bam"
    outbam2 = tmp_path / "output2.bam"

    # both VCFs contain the same positions, but chromosomes are sorted differently
    run_haplotag(
        variant_file="tests/data/haplotag.large.vcf.gz",
        alignment_file="tests/data/haplotag.large.bam",
        output=outbam1,
    )
    run_haplotag(
        variant_file="tests/data/haplotag.large.2.vcf.gz",
        alignment_file="tests/data/haplotag.large.bam",
        output=outbam2,
    )
    for a1, a2 in zip(pysam.AlignmentFile(outbam1), pysam.AlignmentFile(outbam2)):
        assert a1.query_name == a2.query_name
        if a1.has_tag("HP"):
            assert a2.has_tag("HP")
            assert a1.get_tag("HP") == a2.get_tag("HP")


def test_haplotag_10X(tmp_path):
    outbam = tmp_path / "output.bam"
    run_haplotag(
        variant_file="tests/data/haplotag.10X.vcf.gz",
        alignment_file="tests/data/haplotag.10X.bam",
        output=outbam,
    )
    # map BX tag --> readlist
    bx_tag_to_readlist = defaultdict(list)
    for alignment in pysam.AlignmentFile(outbam):
        if alignment.has_tag("BX") and alignment.has_tag("HP"):
            bx_tag_to_readlist[alignment.get_tag("BX")].append(alignment)
    # reads having same BX tag need to be assigned to same haplotype
    for tag in bx_tag_to_readlist.keys():
        haplotype = bx_tag_to_readlist[tag][0].get_tag("HP")
        for read in bx_tag_to_readlist[tag]:
            assert haplotype == read.get_tag("HP")


def test_haplotag_10X_2(tmp_path):
    outbam = tmp_path / "output.bam"
    run_haplotag(
        variant_file="tests/data/haplotag.10X_2.vcf.gz",
        alignment_file="tests/data/haplotag.10X.bam",
        output=outbam,
    )
    for a1, a2 in zip(
        pysam.AlignmentFile("tests/data/haplotag.10X.bam"), pysam.AlignmentFile(outbam)
    ):
        assert a1.query_name == a2.query_name
        if a1.has_tag("HP") and a2.has_tag("HP"):
            assert a1.get_tag("HP") == a2.get_tag("HP")


def test_haplotag_10X_ignore_linked_read(tmp_path):
    outbam_links = tmp_path / "with_links.bam"
    outbam_nolinks = tmp_path / "no_links.bam"
    run_haplotag(
        variant_file="tests/data/haplotag.10X.vcf.gz",
        alignment_file="tests/data/haplotag.10X_3.bam",
        output=outbam_links,
    )
    run_haplotag(
        variant_file="tests/data/haplotag.10X.vcf.gz",
        alignment_file="tests/data/haplotag.10X_3.bam",
        output=outbam_nolinks,
        ignore_linked_read=True,
    )
    expected_links = {"read1": [1, 4], "read2": [1, 4], "read3": [1, 11], "read4": [1, 11]}
    expected_no_links = {"read1": [2, 66], "read2": [1, 70], "read3": [2, 55], "read4": [1, 66]}
    for a1, a2 in zip(pysam.AlignmentFile(outbam_links), pysam.AlignmentFile(outbam_nolinks)):
        assert a1.query_name == a2.query_name
        name = a1.query_name
        if name == "read5":
            # read5 assigned according to other reads with same BX tag
            assert a1.has_tag("HP")
            assert a1.get_tag("HP") == 1
            # using --ignore-linked-read, read5 must be untagged
            assert not a2.has_tag("HP")
        else:
            assert a1.get_tag("HP") == expected_links[name][0]
            assert a1.get_tag("PC") == expected_links[name][1]
            assert a2.get_tag("HP") == expected_no_links[name][0]
            assert a2.get_tag("PC") == expected_no_links[name][1]


def test_haplotag_supplementary(tmp_path):
    # test --tag-supplementary option which assigns supplementary
    # reads to haplotypes based on the tag of their primary alignment.
    outbam1 = tmp_path / "supp-untagged.bam"
    outbam2 = tmp_path / "supp-tagged.bam"
    run_haplotag(
        variant_file="tests/data/haplotag.supplementary.vcf.gz",
        alignment_file="tests/data/haplotag.supplementary.bam",
        output=outbam1,
        ignore_read_groups=True,
    )
    run_haplotag(
        variant_file="tests/data/haplotag.supplementary.vcf.gz",
        alignment_file="tests/data/haplotag.supplementary.bam",
        output=outbam2,
        tag_supplementary=True,
        ignore_read_groups=True,
    )
    # map name->haplotype
    primary_to_tag = {}
    supplementary_to_tag = {}
    for a1, a2 in zip(pysam.AlignmentFile(outbam1), pysam.AlignmentFile(outbam2)):
        assert a1.query_name == a2.query_name
        if a1.has_tag("HP") and a2.has_tag("HP"):
            assert a1.get_tag("HP") == a2.get_tag("HP")
            assert not a1.is_supplementary
        if a2.has_tag("HP"):
            tag = a2.get_tag("HP")
            if a2.is_supplementary:
                supplementary_to_tag[a2.query_name] = tag
            else:
                primary_to_tag[a2.query_name] = tag
    # check if supplementary and primary tags agree
    assert len(primary_to_tag.keys()) == len(supplementary_to_tag.keys()) == 3
    for r, t in supplementary_to_tag.items():
        assert r in primary_to_tag
        primary_tag = primary_to_tag[r]
        assert t == primary_tag


def test_haplotag_regions(tmp_path):
    outbam1 = tmp_path / "output1.bam"
    outbam2 = tmp_path / "output2.bam"
    outlist1 = tmp_path / "list1.tsv"
    outlist2 = tmp_path / "list2.tsv"

    # run haplotag with identical VCF, but once specifying regions
    # output must be identical
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag.bam",
        haplotag_list=outlist1,
        output=outbam1,
        regions=None,
    )
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag.bam",
        haplotag_list=outlist2,
        output=outbam2,
        regions=["chr1"],
    )
    for a1, a2 in zip(pysam.AlignmentFile(outbam1), pysam.AlignmentFile(outbam2)):
        assert a1.query_name == a2.query_name
        if a1.has_tag("HP"):
            assert a2.has_tag("HP")
            assert a1.get_tag("HP") == a2.get_tag("HP")
    for n, (line1, line2) in enumerate(zip(open(outlist1), open(outlist2))):
        assert line1 == line2
    assert n == 20


def test_haplotag_nonexisting_region():
    with pytest.raises(ValueError):
        run_haplotag(
            variant_file="tests/data/haplotag_1.vcf.gz",
            alignment_file="tests/data/haplotag.bam",
            haplotag_list=None,
            output=None,
            regions=["chr2"],
        )


def test_haplotag_selected_regions(tmp_path):
    start1 = 1054025
    end1 = 1069500
    start2 = 1075700
    outbam = tmp_path / "output.bam"
    outlist = tmp_path / "haplolist.tsv"
    run_haplotag(
        variant_file="tests/data/haplotag_1.vcf.gz",
        alignment_file="tests/data/haplotag.bam",
        haplotag_list=outlist,
        output=outbam,
        regions=["chr1:{}-{}".format(start1, end1), "chr1:{}".format(start2)],
    )

    var_region1 = set()
    var_region2 = set()
    unphased_variants = [1074910, 1075707, 1075715]
    with pysam.VariantFile("tests/data/haplotag_1.vcf.gz", "rb") as vcf:
        for variant in vcf:
            if variant.pos in unphased_variants:
                continue
            if start1 <= variant.start <= end1:
                var_region1.add(variant.start)
            elif start2 <= variant.start:
                var_region2.add(variant.start)
            else:
                pass
    # sanity check:
    # there are no variants in the VCF
    # overlapping region 1
    assert len(var_region1) == 0

    with pysam.AlignmentFile(outbam, "rb") as test_bam:
        # Since not all variants from the VCF are selected,
        # count how many variants are overlapping the read.
        # If more than 1 overlap, read must be phased / have HP tag
        for aln in test_bam:
            num_ovl = sum([int(aln.reference_start <= v <= aln.reference_end) for v in var_region2])
            if num_ovl > 1:
                assert aln.has_tag("HP")


def test_cram_output(tmp_path):
    outcram = tmp_path / "output.cram"
    run_haplotag(
        variant_file="tests/data/pacbio/phased.vcf.gz",
        alignment_file="tests/data/pacbio/pacbio.bam",
        reference="tests/data/pacbio/reference.fasta",
        output=outcram,
    )
    with pysam.AlignmentFile(outcram) as f:
        assert f.is_cram


def test_haplotag_unmapped_reads(tmp_path):
    outbam = tmp_path / "output.bam"
    run_haplotag(
        variant_file="tests/data/haplotag.10X.vcf.gz",
        alignment_file="tests/data/unmapped.bam",
        output=outbam,
    )
    pysam.index(str(outbam))
    with pysam.AlignmentFile(outbam) as af:
        alignments = list(af.fetch(until_eof=True))
    assert len(alignments) == 6
    assert not alignments[4].is_unmapped
    assert alignments[5].is_unmapped



import os.path

from pytest import raises
from whatshap.utils import (
    detect_file_format,
    IndexedFasta,
    FastaNotIndexedError,
    Region,
    InvalidRegion,
)


def test_detect_alignment_file_format():
    assert detect_file_format("tests/data/oneread.bam") == "BAM"
    assert detect_file_format("tests/data/oneread.cram") == "CRAM"
    assert detect_file_format("tests/data/onevariant.vcf") == "VCF"
    assert detect_file_format("tests/data/onevariant.vcf.gz") == "VCF"


def test_detect_ped_file_format():
    assert detect_file_format("tests/data/pedigree.ped") is None


def test_missing_fai():
    assert not os.path.exists("tests/data/not-indexed.fasta.fai")
    with raises(FastaNotIndexedError):
        IndexedFasta("tests/data/not-indexed.fasta")


def test_region_start_greater_than_end():
    with raises(InvalidRegion):
        Region.parse("chr1:500-200")
    with raises(InvalidRegion):
        Region.parse("chr1:500-200:17")
    with raises(InvalidRegion):
        Region.parse("chr1:a-b")



__all__ = ["__version__"]

from ._version import version as __version__


def _load_cpp_classes():
    import sys
    import os

    flags = sys.getdlopenflags()
    sys.setdlopenflags(os.RTLD_GLOBAL | os.RTLD_NOW)
    import whatshap.core  # noqa
    import whatshap.polyphase_solver  # noqa

    sys.setdlopenflags(flags)


_load_cpp_classes()



import os
import pysam
import math

import pytest
from pysam import VariantFile
from whatshap.cli.genotype import run_genotype
from whatshap.cli import CommandLineError
from whatshap.vcf import VcfReader

trio_bamfile = "tests/data/trio.pacbio.bam"
trio_merged_bamfile = "tests/data/trio-merged-blocks.bam"
trio_paired_end_bamfile = "tests/data/paired_end.sorted.bam"
ped_samples_bamfile = "tests/data/ped_samples.bam"
recombination_breaks_bamfile = "tests/data/recombination_breaks.sorted.bam"
quartet2_bamfile = "tests/data/quartet2.bam"
short_bamfile = "tests/data/short-genome/short.bam"
indels_bamfile = "tests/data/indels.bam"

bam_files = [
    trio_bamfile,
    trio_merged_bamfile,
    trio_paired_end_bamfile,
    recombination_breaks_bamfile,
    quartet2_bamfile,
    short_bamfile,
    indels_bamfile,
]


def setup_module():
    # This function is run once for this module
    for bam_path in bam_files:
        assert bam_path.endswith(".bam")
        sam_path = bam_path[:-4] + ".sam"
        pysam.view(sam_path, "-b", "-o", bam_path, catch_stdout=False)
        pysam.index(bam_path, catch_stdout=False)


def test_one_variant():
    run_genotype(
        phase_input_files=["tests/data/oneread.bam"],
        variant_file="tests/data/onevariant.vcf",
        output="/dev/null",
    )


def test_default_output():
    """Output to stdout"""
    run_genotype(
        phase_input_files=["tests/data/oneread.bam"], variant_file="tests/data/onevariant.vcf"
    )


def test_bam_without_readgroup():
    run_genotype(
        phase_input_files=["tests/data/no-readgroup.bam"],
        variant_file="tests/data/onevariant.vcf",
        output="/dev/null",
        ignore_read_groups=True,
    )


def test_requested_sample_not_found():
    with pytest.raises(CommandLineError):
        run_genotype(
            phase_input_files=["tests/data/oneread.bam"],
            variant_file="tests/data/onevariant.vcf",
            output="/dev/null",
            samples=["DOES_NOT_EXIST"],
        )


def test_with_reference():
    run_genotype(
        phase_input_files=["tests/data/pacbio/pacbio.bam"],
        variant_file="tests/data/pacbio/variants.vcf",
        reference="tests/data/pacbio/reference.fasta",
    )


@pytest.mark.parametrize("priors", [True, False])
def test_no_indels(tmpdir, priors):
    prioroutput = str(tmpdir.join("priors.vcf")) if priors else None
    outvcf = str(tmpdir.join("output_gl.vcf"))
    run_genotype(
        phase_input_files=["tests/data/pacbio/pacbio.bam"],
        variant_file="tests/data/pacbio/variants.vcf",
        reference="tests/data/pacbio/reference.fasta",
        output=outvcf,
        indels=False,
        nopriors=not priors,
        prioroutput=prioroutput,
    )
    result_vcfs = [outvcf]
    if priors:
        result_vcfs.append(prioroutput)

    # make sure indels not genotyped (also in priors.vcf if computed)
    for o_vcf in result_vcfs:
        vcf_reader = VariantFile(o_vcf)
        default_l = math.log10(1 / 3.0)

        for record in vcf_reader:
            if record.alts is None:
                for call in record.samples.values():
                    assert set(call) == {"GT"}
            elif len(record.alts[0]) != len(record.ref):
                for call in record.samples.values():
                    for v in call["GL"]:
                        assert pytest.approx(default_l) == v


def test_multiallelic(tmpdir):
    outvcf = str(tmpdir.join("output_multi.vcf"))
    run_genotype(
        phase_input_files=["tests/data/pacbio/pacbio.bam"],
        variant_file="tests/data/multiallelic.vcf",
        reference="tests/data/pacbio/reference.fasta",
        output=outvcf,
        indels=False,
    )
    vcf_reader = VariantFile(outvcf)
    for record in vcf_reader:
        n_alleles = len(record.alts) + 1
        if n_alleles > 1:
            for call in record.samples.values():
                assert len(call["GL"]) == ((n_alleles + 1) * n_alleles) / 2


def likeliest_genotype(a, b, c, thres):
    prob_a = 10**a
    prob_b = 10**b
    prob_c = 10**c

    prob = sorted([(prob_a, 0), (prob_b, 1), (prob_c, 2)])

    if prob[2][0] > prob[1][0] and prob[2][0] > thres:
        return prob[2][1]
    else:
        return None


@pytest.mark.parametrize("threshold", [0, 2, 3, 6, 13, 50])
def test_gt_quality_threshold(threshold, tmpdir):
    thres = 1 - 10 ** (-threshold / 10.0)

    out_vcf = str(tmpdir.join("out.vcf"))
    priors_vcf = str(tmpdir.join("priors.vcf"))
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        output=out_vcf,
        gt_qual_threshold=threshold,
        indels=False,
        prioroutput=priors_vcf,
    )

    for path in [out_vcf, priors_vcf]:
        for record in VariantFile(path):
            for call in record.samples.values():
                likelihoods = call["GL"]
                genotype = call["GT"]
                if genotype == (None,):
                    genotype = None
                else:
                    genotype = genotype[0] + genotype[1]
                gt = likeliest_genotype(likelihoods[0], likelihoods[1], likelihoods[2], thres)
                # print(likelihoods[0], likelihoods[1], likelihoods[2], gt, genotype, thres)
                assert gt == genotype


def test_genotyping_one_of_three_individuals(tmp_path):
    outvcf = tmp_path / "output.vcf"
    outpriors = tmp_path / "priors.vcf"
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        output=outvcf,
        samples=["HG003"],
        prioroutput=outpriors,
    )

    for outfile in [outvcf, outpriors]:
        assert os.path.isfile(outfile)

        tables = list(VcfReader(outfile, phases=True, genotype_likelihoods=True))
        assert len(tables) == 1
        table = tables[0]
        assert table.chromosome == "1"
        assert len(table.variants) == 5
        assert table.samples == ["HG004", "HG003", "HG002"]

        # there should be no genotype predictions for HG003/HG002
        default_l = math.log10(1 / 3.0)
        for l in [
            table.genotype_likelihoods_of("HG002"),
            table.genotype_likelihoods_of("HG004"),
        ]:
            for var in l:
                for v in var.log10_probs():
                    assert pytest.approx(default_l) == v


def test_use_ped_samples(tmp_path):
    outvcf = tmp_path / "output_ped_samples.vcf"
    run_genotype(
        phase_input_files=[ped_samples_bamfile],
        variant_file="tests/data/ped_samples.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        use_ped_samples=True,
    )
    assert os.path.isfile(outvcf)

    tables = list(VcfReader(outvcf, phases=True, genotype_likelihoods=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002", "orphan"]

    default_l = math.log10(1 / 3.0)
    for var in table.genotype_likelihoods_of("orphan"):
        for v in var.log10_probs():
            assert pytest.approx(default_l) == v


@pytest.mark.parametrize(
    "sample_set",
    [["HG002"], ["HG003"], ["HG004"], ["HG002", "HG003"], ["HG002", "HG004"], ["HG003", "HG004"]],
)
def test_ped_sample(sample_set, tmp_path):
    # running with --ped and --sample on subset of trio,
    # should give same results as running with only --sample
    # the trio information should be ignored
    outvcf1 = tmp_path / "output1.vcf"
    outvcf2 = tmp_path / "output2.vcf"
    run_genotype(
        phase_input_files=[ped_samples_bamfile],
        variant_file="tests/data/ped_samples.vcf",
        output=outvcf1,
        ped="tests/data/trio.ped",
        samples=sample_set,
    )
    run_genotype(
        phase_input_files=[ped_samples_bamfile],
        variant_file="tests/data/ped_samples.vcf",
        output=outvcf2,
        samples=sample_set,
    )
    assert os.path.isfile(outvcf1)
    assert os.path.isfile(outvcf2)
    tables1 = list(VcfReader(outvcf1, phases=True, genotype_likelihoods=True))
    tables2 = list(VcfReader(outvcf2, phases=True, genotype_likelihoods=True))
    assert (len(tables1) == 1) and (len(tables2) == 1)
    table1, table2 = tables1[0], tables2[0]

    for individual in sample_set:
        for var1, var2 in zip(
            table1.genotype_likelihoods_of(individual), table2.genotype_likelihoods_of(individual)
        ):
            print(var1, var2)
            assert var1.log10_probs() == var2.log10_probs()


def test_genotyping_trio(tmp_path):
    outvcf = tmp_path / "output.vcf"
    outpriors = tmp_path / "priors.vcf"
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        prioroutput=outpriors,
    )

    for outfile in [outvcf, outpriors]:
        assert os.path.isfile(outfile)

        tables = list(VcfReader(outfile, phases=True))
        assert len(tables) == 1
        table = tables[0]
        assert table.chromosome == "1"
        assert len(table.variants) == 5
        assert table.samples == ["HG004", "HG003", "HG002"]


@pytest.mark.parametrize("chromosome", ["1", "2"])
def test_genotyping_specific_chromosome(chromosome, tmp_path):
    outvcf = tmp_path / "output.vcf"
    outpriors = tmp_path / "priors.vcf"
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio-two-chromosomes.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        chromosomes=[chromosome],
        prioroutput=outpriors,
    )
    for outfile in [outvcf, outpriors]:
        assert os.path.isfile(outfile)
        tables = list(VcfReader(outfile, genotype_likelihoods=True))
        assert len(tables) == 2
        for table in tables:
            assert len(table.variants) == 5
            assert table.samples == ["HG004", "HG003", "HG002"]

        index = 0
        if chromosome == "1":
            index = 1

        # should be no genotype likelihoods for skipped chromosomes
        for s in tables[index].samples:
            assert tables[index].genotype_likelihoods_of(s) == [None] * 5
            assert tables[not index].genotype_likelihoods_of(s) != [None] * 5


def test_genotype_likelihoods_given(tmp_path):
    outvcf = tmp_path / "output_gl.vcf"
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio_genotype_likelihoods.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
    )
    assert os.path.isfile(outvcf)
    tables = list(VcfReader(outvcf, phases=True, genotype_likelihoods=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 5
    assert table.samples == ["HG004", "HG003", "HG002"]

    # check if PL likelihoods (that were present before) are deleted
    vcf_reader = VariantFile(outvcf)
    # print(list(vcf_reader.samples), outvcf)
    for record in vcf_reader:
        for call in record.samples.values():
            PL = call.get("PL", None)
            GL = call.get("GL", None)
            print("GL:", GL, "PL:", PL)
            assert PL == (None, None, None)
            assert GL is not None


# GL field was already present, make sure it is replaced by new likelihoods
def test_genotype_log_likelihoods_given(tmp_path):
    outvcf = tmp_path / "output_gl_log.vcf"
    outpriors = tmp_path / "priors.vcf"
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio_genotype_log_likelihoods.vcf",
        output=outvcf,
        ped="tests/data/trio.ped",
        genmap="tests/data/trio.map",
        gt_qual_threshold=0,
        prioroutput=outpriors,
    )
    for outfile in [outvcf, outpriors]:
        assert os.path.isfile(outfile)
        tables = list(VcfReader(outfile, phases=True, genotype_likelihoods=True))
        assert len(tables) == 1
        table = tables[0]
        assert table.chromosome == "1"
        assert len(table.variants) == 5
        assert table.samples == ["HG004", "HG003", "HG002"]

        # check if GL likelihoods were replaced
        vcf_reader = VariantFile(outfile)
        print(list(vcf_reader.header.samples), outfile)
        for record in vcf_reader:
            for call in record.samples.values():
                GL = call.get("GL", None)
                GQ = call.get("GQ", None)
                print("GL:", GL, "GQ", GQ)
                assert GL != [-1, -1, -1]
                assert GQ != 100


def test_empty_format_field(tmp_path):
    outvcf = tmp_path / "output_empty_format.vcf"
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/empty_format.vcf",
        output=outvcf,
        gt_qual_threshold=0,
    )

    # check if sample fields now contain information
    assert os.path.isfile(outvcf)
    vcf_reader = VariantFile(outvcf)
    for record in vcf_reader:
        for sample, call in record.samples.items():
            assert set(call) == {"GT", "GL", "GQ"}


def test_phase_trio_paired_end_reads(tmp_path):
    outvcf = tmp_path / "output-paired_end.vcf"
    run_genotype(
        phase_input_files=[trio_paired_end_bamfile],
        variant_file="tests/data/paired_end.sorted.vcf",
        output=outvcf,
        ped="tests/data/trio_paired_end.ped",
        genmap="tests/data/trio.map",
    )
    assert os.path.isfile(outvcf)
    tables = list(VcfReader(outvcf, phases=True))
    assert len(tables) == 1
    table = tables[0]
    assert table.chromosome == "1"
    assert len(table.variants) == 3
    assert table.samples == ["mother", "father", "child"]


def test_wrong_chromosome(tmp_path):
    outvcf = tmp_path / "output.vcf"
    with pytest.raises(CommandLineError):
        run_genotype(
            phase_input_files=[short_bamfile],
            ignore_read_groups=True,
            variant_file="tests/data/short-genome/wrongchromosome.vcf",
            output=outvcf,
        )


def extract_likelihoods(record):
    return [10.0**gl for gl in record.samples[0]["GL"]]


@pytest.mark.parametrize("constant", [0.1, 0.2, 0.3, 0.5, 0.7, 1, 2, 5, 10, 20, 100])
def test_adding_constant(constant, tmpdir):
    priors_raw_vcf = str(tmpdir.join("output.raw_priors.vcf"))
    outvcf_raw_vcf = str(tmpdir.join("output_raw.vcf"))
    priors_const_vcf = str(tmpdir.join("output.const_priors.vcf"))
    outvcf_const_vcf = str(tmpdir.join("output_raw.vcf"))

    # run genotyping without adding constant to priors
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        prioroutput=priors_raw_vcf,
        output=outvcf_raw_vcf,
        indels=False,
    )

    # run genotyping with modified priors
    run_genotype(
        phase_input_files=[trio_bamfile],
        variant_file="tests/data/trio.vcf",
        prioroutput=priors_const_vcf,
        output=outvcf_const_vcf,
        indels=False,
        constant=constant,
    )

    # check if priors were modified properly
    with pysam.VariantFile(priors_raw_vcf) as f:
        records_raw = list(f)
    with pysam.VariantFile(priors_const_vcf) as f:
        records_const = list(f)

    assert len(records_raw) == len(records_const)

    for record_raw, record_const in zip(records_raw, records_const):
        likelihoods_raw = extract_likelihoods(record_raw)
        likelihoods_const = extract_likelihoods(record_const)
        norm_sum = likelihoods_raw[0] + likelihoods_raw[1] + likelihoods_raw[2] + 3.0 * constant
        # print(float(likelihoods_const[0]), float((likelihoods_raw[0] + constant)/norm_sum))
        # print(float(likelihoods_const[1]), float((likelihoods_raw[1] + constant)/norm_sum))
        # print(float(likelihoods_const[2]), float((likelihoods_raw[2] + constant)/norm_sum))

        for j in range(3):
            assert (
                pytest.approx(likelihoods_const[j], 1e-5)
                == (likelihoods_raw[j] + constant) / norm_sum
            )



"""
Print the square of a number

This is an example subcommand. The first line above is a short summary description
of what the command does. It is shown when running whatshap --help. Make it short!

To add a new subcommand:

* Copy this module to a new file and adjust it to your needs. The name of the
  module is identical to the subcommand name.
* Add the name of this module to the ``COMMANDS`` list in ``whatshap/__main__.py``.

The module must have add_arguments() and main() methods. The validate() method is optional and
can be used to validate the command-line options (use parser.error to raise an
error).
"""
import logging

logger = logging.getLogger(__name__)


def add_arguments(parser):
    add = parser.add_argument
    add("number", type=int, help="The number to square (at least 10)")


def validate(args, parser):
    if args.number <= 10:
        parser.error("Sorry, this is too simple!")


def main(args):
    print("{} * {} = {}".format(args.number, args.number, args.number * args.number))



"""
Test phasing of pedigrees (PedMEC algorithm)
"""
from collections import defaultdict
from pytest import raises
from whatshap.core import (
    PedigreeDPTable,
    ReadSet,
    Pedigree,
    NumericSampleIds,
    PhredGenotypeLikelihoods,
)
from whatshap.pedigree import centimorgen_to_phred
from whatshap.testhelpers import string_to_readset_pedigree, canonic_index_list_to_biallelic_gt_list


def phase_pedigree(reads, recombcost, pedigree, distrust_genotypes=False, positions=None):
    rs = string_to_readset_pedigree(reads)
    dp_table = PedigreeDPTable(rs, recombcost, pedigree, distrust_genotypes, positions)
    superreads_list, transmission_vector = dp_table.get_super_reads()
    cost = dp_table.get_optimal_cost()
    for superreads in superreads_list:
        for sr in superreads:
            print(sr)
    print("Cost:", dp_table.get_optimal_cost())
    print("Transmission vector:", transmission_vector)
    print("Partition:", dp_table.get_optimal_partitioning())
    return superreads_list, transmission_vector, cost


def assert_haplotypes(superreads_list, all_expected_haplotypes, length):
    for superreads, expected_haplotypes in zip(superreads_list, all_expected_haplotypes):
        assert len(superreads) == 2
        assert len(superreads[0]) == len(superreads[1]) == length
        haplotypes = tuple(sorted("".join(str(v.allele) for v in sr) for sr in superreads))
        assert (haplotypes == (expected_haplotypes[0], expected_haplotypes[1])) or (
            haplotypes == (expected_haplotypes[1], expected_haplotypes[0])
        )


def assert_trio_allele_order(superreads_list, transmission_vector, nr_of_positions):
    # assume superreads_list contains superreads for father, mother, child (in that order!)
    assert len(superreads_list) == 3
    father = superreads_list[0]
    mother = superreads_list[1]
    child = superreads_list[2]

    for pos in range(nr_of_positions):
        transmission_value = transmission_vector[pos]
        paternal_transmission = transmission_value % 2
        maternal_transmission = transmission_value // 2
        paternal_allele = father[not paternal_transmission][pos].allele
        maternal_allele = mother[not maternal_transmission][pos].allele
        child_allele_p = child[0][pos].allele
        child_allele_m = child[1][pos].allele
        print(
            "position: ",
            pos,
            "paternal allele: ",
            paternal_allele,
            "maternal allele",
            maternal_allele,
            "child genotype: ",
            child_allele_p,
            child_allele_m,
        )
        assert paternal_allele == child_allele_p
        assert maternal_allele == child_allele_m


def get_trio_transmission_vectors(transmission_vector, nr_of_trios):
    trio_transmission_vectors = defaultdict(list)
    for transmission_value in transmission_vector:
        for trio in range(nr_of_trios):
            value = transmission_value % 4
            transmission_value = transmission_value // 4
            trio_transmission_vectors[trio].append(value)
    return trio_transmission_vectors


def test_phase_empty_trio():
    rs = ReadSet()
    recombcost = []
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individual0", [])
    pedigree.add_individual("individual1", [])
    pedigree.add_individual("individual2", [])
    pedigree.add_relationship("individual0", "individual1", "individual2")
    dp_table = PedigreeDPTable(rs, recombcost, pedigree)
    ((superreadsm, superreadsf, superreadsc), transmission_vector) = dp_table.get_super_reads()


def test_phase_trio1():
    reads = """
      A 111
      A 010
      A 110
      B 001
      B 110
      B 101
      C 001
      C 010
      C 010
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individual0", canonic_index_list_to_biallelic_gt_list([1, 2, 1]))
    pedigree.add_individual("individual1", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_individual("individual2", canonic_index_list_to_biallelic_gt_list([0, 1, 1]))
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    assert cost == 2
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [("111", "010"), ("001", "110"), ("010", "001")]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 3)
    assert_trio_allele_order(superreads_list, transmission_vector, 3)


def test_phase_trio2():
    reads = """
      A 00
      A 00
      B 11
      B 11
      C 11
      C 00
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individual0", canonic_index_list_to_biallelic_gt_list([2, 2]))
    pedigree.add_individual("individual1", canonic_index_list_to_biallelic_gt_list([0, 0]))
    pedigree.add_individual("individual2", canonic_index_list_to_biallelic_gt_list([1, 1]))
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    assert cost == 8
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [("11", "11"), ("00", "00"), ("00", "11")]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 2)
    assert_trio_allele_order(superreads_list, transmission_vector, 2)


def test_phase_trio3():
    reads = """
      A 1111
      B 1010
      C 111000
      C 010101
      B 0101
      A  0000
      B  1010
      C  1010
      C  1100
      A   0000
      A   1111
      B   1010
      B    010
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual(
        "individual0", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1])
    )
    pedigree.add_individual(
        "individual1", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1])
    )
    pedigree.add_individual(
        "individual2", canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1, 0, 1])
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [3, 3, 3, 4, 3, 3]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    assert cost == 4
    assert transmission_vector in (
        [0, 0, 0, 1, 1, 1],
        [1, 1, 1, 0, 0, 0],
        [2, 2, 2, 3, 3, 3],
        [3, 3, 3, 2, 2, 2],
    )
    all_expected_haplotypes = [
        ("111111", "000000"),
        ("010101", "101010"),
        ("111000", "010101"),
    ]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 6)
    assert_trio_allele_order(superreads_list, transmission_vector, 6)


def test_phase_trio4():
    reads = """
      B 101
      B 101
      B 101
      A 111
      A 111
      A 111
      C 111
      C 111
      C 111
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individual0", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_individual("individual1", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_individual("individual2", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [1, 1, 1]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    assert cost == 2
    assert transmission_vector in ([0, 2, 0], [2, 0, 2], [1, 3, 1], [3, 1, 3])
    all_expected_haplotypes = [("111", "000"), ("101", "010"), ("111", "000")]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 3)
    assert_trio_allele_order(superreads_list, transmission_vector, 3)


def test_phase_trio5():
    reads = """
      B 101
      B 101
      B 101
      A 111
      A 111
      A 111
      C 111
      C 111
      C 111
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individual0", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_individual("individual1", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_individual("individual2", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [2, 2, 2]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    assert cost == 3
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [("111", "000"), ("111", "000"), ("111", "000")]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 3)
    assert_trio_allele_order(superreads_list, transmission_vector, 3)


def test_phase_trio_pure_genetic():
    reads = ""
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individual0", canonic_index_list_to_biallelic_gt_list([2, 1, 1, 0]))
    pedigree.add_individual("individual1", canonic_index_list_to_biallelic_gt_list([1, 2, 2, 1]))
    pedigree.add_individual("individual2", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 0]))
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [2, 2, 2]
    superreads_list, transmission_vector, cost = phase_pedigree(
        reads, recombcost, pedigree, positions=[10, 20, 30, 40]
    )
    assert cost == 0
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [("1110", "1000"), ("1111", "0110"), ("1000", "0110")]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 4)
    assert_trio_allele_order(superreads_list, transmission_vector, 4)


def test_phase_doubletrio_pure_genetic():
    reads = ""
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individualA", canonic_index_list_to_biallelic_gt_list([1, 2, 1, 0]))
    pedigree.add_individual("individualB", canonic_index_list_to_biallelic_gt_list([1, 0, 1, 1]))
    pedigree.add_individual("individualC", canonic_index_list_to_biallelic_gt_list([2, 1, 1, 0]))
    pedigree.add_individual("individualD", canonic_index_list_to_biallelic_gt_list([1, 2, 2, 1]))
    pedigree.add_individual("individualE", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 0]))
    pedigree.add_relationship("individualA", "individualB", "individualC")
    pedigree.add_relationship("individualC", "individualD", "individualE")
    recombcost = [2, 2, 2]
    superreads_list, transmission_vector, cost = phase_pedigree(
        reads, recombcost, pedigree, positions=[10, 20, 30, 40]
    )
    assert cost == 0
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [
        ("0100", "1110"),
        ("0011", "1000"),
        ("1110", "1000"),
        ("1111", "0110"),
        ("1000", "0110"),
    ]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 4)
    trio_transmission_vectors = get_trio_transmission_vectors(transmission_vector, 4)
    assert_trio_allele_order(superreads_list[:3], trio_transmission_vectors[0], 4)
    assert_trio_allele_order(superreads_list[2:], trio_transmission_vectors[1], 4)


def test_phase_quartet1():
    reads = """
      A 111
      A 010
      A 110
      B 001
      B 110
      B 101
      C 001
      C 010
      C 010
      D 001
      D 010
      D 010
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual("individual0", canonic_index_list_to_biallelic_gt_list([1, 2, 1]))
    pedigree.add_individual("individual1", canonic_index_list_to_biallelic_gt_list([1, 1, 1]))
    pedigree.add_individual("individual2", canonic_index_list_to_biallelic_gt_list([0, 1, 1]))
    pedigree.add_individual("individual3", canonic_index_list_to_biallelic_gt_list([0, 1, 1]))
    pedigree.add_relationship("individual0", "individual1", "individual2")
    pedigree.add_relationship("individual0", "individual1", "individual3")
    recombcost = [10, 10, 10]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    assert cost == 2
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [
        ("111", "010"),
        ("001", "110"),
        ("001", "010"),
        ("001", "010"),
    ]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 3)
    trio_transmission_vectors = get_trio_transmission_vectors(transmission_vector, 3)
    assert_trio_allele_order(superreads_list[:3], trio_transmission_vectors[0], 3)
    assert_trio_allele_order(
        [superreads_list[0], superreads_list[1], superreads_list[3]],
        trio_transmission_vectors[1],
        3,
    )


def test_phase_quartet2():
    reads = """
      A 111111
      A 000000
      B 010101
      B 101010
      C 000000
      C 010101
      D 000000
      D 010101
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual(
        "individual0", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1])
    )
    pedigree.add_individual(
        "individual1", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1])
    )
    pedigree.add_individual(
        "individual2", canonic_index_list_to_biallelic_gt_list([0, 1, 0, 1, 0, 1])
    )
    pedigree.add_individual(
        "individual3", canonic_index_list_to_biallelic_gt_list([0, 1, 0, 1, 0, 1])
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    pedigree.add_relationship("individual0", "individual1", "individual3")
    recombcost = [3, 3, 3, 3, 3, 3]

    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    assert cost == 0
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [
        ("111111", "000000"),
        ("010101", "101010"),
        ("000000", "010101"),
        ("000000", "010101"),
    ]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 6)
    trio_transmission_vectors = get_trio_transmission_vectors(transmission_vector, 6)
    assert_trio_allele_order(superreads_list[:3], trio_transmission_vectors[0], 6)
    assert_trio_allele_order(
        [superreads_list[0], superreads_list[1], superreads_list[3]],
        trio_transmission_vectors[1],
        6,
    )


def test_phase_quartet3():
    reads = """
      A 1111
      A 0000
      B 1010
      C 111000
      C 010101
      D 000000
      D 010
      B 0101
      C  1100
      D  10010
      A   0000
      A   1111
      B   1010
      B   0101
    """
    pedigree = Pedigree(NumericSampleIds())
    pedigree.add_individual(
        "individual0", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1])
    )
    pedigree.add_individual(
        "individual1", canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1])
    )
    pedigree.add_individual(
        "individual2", canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1, 0, 1])
    )
    pedigree.add_individual(
        "individual3", canonic_index_list_to_biallelic_gt_list([0, 1, 0, 0, 1, 0])
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    pedigree.add_relationship("individual0", "individual1", "individual3")
    recombcost = [3, 3, 3, 4, 3, 3]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree)
    print(cost)
    print(transmission_vector)
    assert cost == 8
    # TODO: expect transmission in both trio relations. Update once transmission vectors
    #       are returned per trio relationship
    # assert transmission_vector in ([0,0,0,1,1,1], [1,1,1,0,0,0], [2,2,2,3,3,3], [3,3,3,2,2,2])
    all_expected_haplotypes = [
        ("111111", "000000"),
        ("010101", "101010"),
        ("111000", "010101"),
        ("000000", "010010"),
    ]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 6)
    trio_transmission_vectors = get_trio_transmission_vectors(transmission_vector, 6)
    assert_trio_allele_order(superreads_list[:3], trio_transmission_vectors[0], 6)
    assert_trio_allele_order(
        [superreads_list[0], superreads_list[1], superreads_list[3]],
        trio_transmission_vectors[1],
        6,
    )


def test_centimorgen_to_phred():
    assert round(centimorgen_to_phred(0.10010013353365396)) == 30
    assert round(centimorgen_to_phred(0.0010000100001343354)) == 50
    assert round(centimorgen_to_phred(1e-38)) == 400


def test_centimorgen_to_phred_zero():
    with raises(ValueError):
        assert centimorgen_to_phred(0)


def test_phase_trio_genotype_likelihoods():
    reads = """
      A 111
      A 010
      A 110
      B 001
      B 110
      B 101
      C 001
      C 010
      C 010
    """
    pedigree = Pedigree(NumericSampleIds())
    genotype_likelihoods_mother = [
        PhredGenotypeLikelihoods([0, 0, 0]),
        PhredGenotypeLikelihoods([0, 0, 1]),
        PhredGenotypeLikelihoods([5, 0, 5]),
    ]
    genotype_likelihoods0 = [PhredGenotypeLikelihoods([0, 0, 0])] * 3
    pedigree.add_individual(
        "individual0",
        canonic_index_list_to_biallelic_gt_list([0, 0, 0]),
        genotype_likelihoods_mother,
    )
    pedigree.add_individual(
        "individual1", canonic_index_list_to_biallelic_gt_list([0, 0, 0]), genotype_likelihoods0
    )
    pedigree.add_individual(
        "individual2", canonic_index_list_to_biallelic_gt_list([0, 0, 0]), genotype_likelihoods0
    )
    pedigree.add_relationship("individual0", "individual1", "individual2")
    recombcost = [10, 10, 10]
    superreads_list, transmission_vector, cost = phase_pedigree(reads, recombcost, pedigree, True)
    assert cost == 3
    assert len(set(transmission_vector)) == 1
    all_expected_haplotypes = [("111", "010"), ("001", "110"), ("001", "010")]
    assert_haplotypes(superreads_list, all_expected_haplotypes, 3)
    assert_trio_allele_order(superreads_list, transmission_vector, 3)



from pytest import raises
from whatshap.graph import Graph, CyclicGraphError


def assert_toposort(tuples):
    graph = Graph()
    for mother, father, child in tuples:
        graph.add_edge(child, mother)
        graph.add_edge(child, father)
    t = graph.toposorted()
    assert len(t) == len(set(t))
    # parents must be before children
    for mother, father, child in tuples:
        assert t.index(mother) < t.index(child)
        assert t.index(father) < t.index(child)


def test_graph():
    tuples = [
        # mother, father, child
        ("mmm", "mmf", "mm"),
        ("mf", "mm", "m"),
        ("m", "f", "c1"),
        ("m", "f", "c2"),
        ("ff", "fm", "f"),
    ]
    assert_toposort(tuples)


def test_charles_ii():
    # source: https://en.wikipedia.org/wiki/Charles_II_of_Spain
    # fields are: child, father, mother
    pedigree = """
    Charles II of Spain, Philip IV of Spain, Mariana of Austria

    Mariana of Austria, Ferdinand III, Maria Anna of Spain

    Philip IV of Spain, Philip III of Spain, Margaret of Austria
    Maria Anna of Spain, Philip III of Spain, Margaret of Austria
    Ferdinand III, Ferdinand II, Maria Anna of Bavaria (1574-1616)

    Philip III of Spain, Philip II of Spain, Anna of Austria (1549-80)
    Margaret of Austria, Charles II Archduke of Austria, Maria Anna of Bavaria
    Ferdinand II, Charles II Archduke of Austria, Maria Anna of Bavaria
    Maria Anna of Bavaria (1574-1616), William V Duke of Bavaria, Renata of Lorraine

    Anna of Austria (1549-80), Maximilian II, Maria of Spain
    Maria Anna of Bavaria, Albert V Duke of Bavaria, Anna of Austria
    William V Duke of Bavaria, Albert V Duke of Bavaria, Anna of Austria
    Renata of Lorraine, Francis I Duke of Lorraine, Christina of Denmark

    Philip II of Spain, Charles V, Isabella of Portugal
    Maria of Spain, Charles V, Isabella of Portugal
    Maximilian II, Ferdinand I, Anna of Bohemia and Hungary
    Charles II Archduke of Austria, Ferdinand I, Anna of Bohemia and Hungary
    Anna of Austria, Ferdinand I, Anna of Bohemia and Hungary
    Christina of Denmark, Christian II of Denmark, Isabella of Austria
    Charles V, Philip I of Castile, Joanna of Castille
    Ferdinand I, Philip I of Castile, Joanna of Castille
    Anna of Bohemia and Hungary, Philip I of Castile, Joanna of Castille
    Isabella of Austria, Philip I of Castile, Joanna of Castille
    """
    individuals = set()
    tuples = []
    for line in pedigree.split("\n"):
        line = line.strip()
        if line == "":
            continue
        child, father, mother = line.split(", ")
        assert child not in individuals, child
        tuples.append((mother, father, child))
    assert_toposort(tuples)


def test_cyclic():
    graph = Graph()
    tuples = [
        # mother, father, child
        ("mmm", "mmf", "mm"),
        ("mf", "mm", "m"),
        ("m", "f", "c1"),
        ("m", "f", "c2"),
        ("ff", "fm", "f"),
        ("c1", "c2", "mmf"),  # this makes it cyclic
    ]
    for mother, father, child in tuples:
        graph.add_edge(child, mother)
        graph.add_edge(child, father)
    with raises(CyclicGraphError):
        graph.toposorted()



from whatshap.core import Pedigree, PhredGenotypeLikelihoods, NumericSampleIds

from whatshap.pedigree import find_recombination, RecombinationEvent
from whatshap.testhelpers import canonic_index_list_to_biallelic_gt_list


def test_pedigree_no_gls():
    ped = Pedigree(NumericSampleIds())
    genotypes1 = canonic_index_list_to_biallelic_gt_list([0, 1, 0, 2])
    genotypes5 = canonic_index_list_to_biallelic_gt_list([1, 2, 2, 0])
    ped.add_individual("sample1", genotypes1)
    assert len(ped) == 1
    assert ped.variant_count == 4, str(ped.variant_count)
    ped.add_individual("sample5", genotypes5)
    assert len(ped) == 2
    assert ped.variant_count == 4, str(ped.variant_count)
    for i in range(ped.variant_count):
        assert ped.genotype("sample1", i) == genotypes1[i]
        assert ped.genotype_likelihoods("sample1", i) is None
        assert ped.genotype("sample5", i) == genotypes5[i]
        assert ped.genotype_likelihoods("sample5", i) is None


def test_pedigree_with_gls():
    ped = Pedigree(NumericSampleIds())
    genotypes1 = canonic_index_list_to_biallelic_gt_list([0, 1, 0, 2])
    gls1 = [
        PhredGenotypeLikelihoods([0, 1, 2]),
        PhredGenotypeLikelihoods([215, 81, 147]),
        PhredGenotypeLikelihoods([199, 49, 253]),
        PhredGenotypeLikelihoods([167, 200, 163]),
    ]
    genotypes5 = canonic_index_list_to_biallelic_gt_list([1, 2, 2, 0])
    gls5 = [
        PhredGenotypeLikelihoods([184, 71, 233]),
        PhredGenotypeLikelihoods([65, 32, 87]),
        PhredGenotypeLikelihoods([28, 215, 131]),
        PhredGenotypeLikelihoods([98, 250, 137]),
    ]
    ped.add_individual("sample1", genotypes1, gls1)
    assert len(ped) == 1
    assert ped.variant_count == 4, str(ped.variant_count)
    ped.add_individual("sample5", genotypes5, gls5)
    assert len(ped) == 2
    assert ped.variant_count == 4, str(ped.variant_count)
    for i in range(ped.variant_count):
        assert ped.genotype("sample1", i) == genotypes1[i]
        assert list(ped.genotype_likelihoods("sample1", i)) == list(gls1[i])
        assert ped.genotype("sample5", i) == genotypes5[i]
        assert list(ped.genotype_likelihoods("sample5", i)) == list(gls5[i])


def test_find_recombination():
    transmission_vector = [0, 0, 1, 1, 0]
    positions = [5303, 5432, 8307, 9000, 9500]
    recombcost = [0, 3, 3, 1, 1]
    components = {5303: 5303, 5432: 5303, 8307: 5303, 9000: 5303, 9500: 5303}
    events = find_recombination(transmission_vector, components, positions, recombcost)
    assert events == [
        RecombinationEvent(
            position1=5432,
            position2=8307,
            transmitted_hap_father1=0,
            transmitted_hap_father2=1,
            transmitted_hap_mother1=0,
            transmitted_hap_mother2=0,
            recombination_cost=3,
        ),
        RecombinationEvent(
            position1=9000,
            position2=9500,
            transmitted_hap_father1=1,
            transmitted_hap_father2=0,
            transmitted_hap_mother1=0,
            transmitted_hap_mother2=0,
            recombination_cost=1,
        ),
    ]



import itertools as it
import logging
from collections import defaultdict
from math import log, exp
from enum import Enum

from whatshap.polyphaseutil import get_position_map

logger = logging.getLogger(__name__)


class ReorderType(Enum):
    NONE = 0
    MULTI = 1
    COLLAPSED = 2
    COLLAPSED_PRE = 3


class ReorderEvent:
    def __init__(self, position, event_type, best_prob, second_prob, haplotypes):
        self.position = position
        self.event_type = event_type
        ratio = best_prob / (best_prob + second_prob) if best_prob + second_prob > 0 else 0
        self.confidence = ratio
        self.haplotypes = sorted(haplotypes[:])


def run_reordering(
    readset,
    clustering,
    path,
    haplotypes,
    block_cut_sensitivity,
    error_rate=0.025,
):
    """
    Main method for the reordering stage of the polyploid phasing algorithm. Input:

    readset -- The fragment matrix to phase
    clustering -- A list of clusters. Each cluster is a list of read ids, indicating which reads it
                  contains. Every read can only be present in one cluster.
    path -- The computed cluster paths from the threading stage
    haplotypes -- The computed haplotype sequences (as numeric alleles) from the threading stage
    block_cut_sensitivity -- Policy how conversative the block cuts have to be done. 0 is one
                             phasing block no matter what, 5 is very short blocks

    In this phase, ambiguous sites are reordered by using the most likely arrangement according to
    the reads. Ambiguous sites (resolved in this order) are:

    1. Two (or more) haplotypes being threaded through the same cluster but not having the same
       allele according to precomputed haplotypes. The ordering of these alleles does not matter
       for the threading stage, but multiple of such positions should be ordered to match the input
       reads.
    2. Two (or more) haplotypes being threaded through the same cluster but only a subset of them
       leaves the cluster. Since haplotypes have a history, it matters which haplotype is selected
       to leave, even though the coverage-based threading does not distinguish. Alternatively, two
       (or more) haplotypes on different clusters switch to a new one simultaneously. Here it again
       matters which haplotype goes to which, but all combinations have been equally good in the
       threading stage. These two types of ambiguities are resolved in ascending genome position
       order.
    3. Based on uncertainty from the previous two steps and the selected block cut sensitivity,
       the clustering must eventually be split into blocks.

    """

    # compute auxiliary data
    index, rev_index = get_position_map(readset)

    # determine snp positions inside clusters
    cwise_snps = find_cluster_snps(path, haplotypes)

    # phase cluster snps
    phase_cluster_snps(
        path, haplotypes, cwise_snps, clustering, readset, index, error_rate, window_size=32
    )

    # select most likely continuation for ambiguous haplotype switches
    events = resolve_ambiguous_switches(
        path, haplotypes, clustering, readset, index, error_rate, window_size=32
    )

    cut_positions, haploid_cuts = compute_cut_positions(path, block_cut_sensitivity, events)

    logger.debug(f"Cut positions: {cut_positions}")
    ploidy = len(haplotypes)
    for i in range(ploidy):
        logger.debug(f"Cut positions on phase {i}: {haploid_cuts[i]}")

    return (cut_positions, haploid_cuts, path, haplotypes)


def find_cluster_snps(path, haplotypes):
    """
    For each cluster, finds the positions, where it has a multiplicity of at least 2 with at least
    2 different alleles. These positions have to be phased within the clusters.
    """
    cwise_snps = defaultdict(list)
    for pos in range(len(path)):
        clusters = set()
        alleles = defaultdict(set)
        for i, cid in enumerate(path[pos]):
            clusters.add(cid)
            alleles[cid].add(haplotypes[i][pos])
        for cid in clusters:
            if len(alleles[cid]) >= 2:
                cwise_snps[cid].append(pos)

    return cwise_snps


def phase_cluster_snps(
    path, haplotypes, cwise_snps, clustering, readset, index, error_rate, window_size
):
    """
    For clusters with multiplicity >= 2 on multiple positions: Bring the alleles of the
    precomputed haplotypes in order by using read information.
    The window size determines how many previous positions are taken into account when resolving
    the next one.
    """
    # sort clusters by position of first haplotype
    collapsed = sorted(
        ((cid, sorted(snps)) for cid, snps in cwise_snps.items()), key=lambda x: x[1][0]
    )

    # iterate cluster-wise
    for cid, snps in collapsed:
        # determine the last window_size many heterozyguous positions on the threads of the first snp
        c_slots = [j for j in range(len(path[snps[0]])) if path[snps[0]][j] == cid]
        het_pos = []
        i, w = snps[0] - 1, 0
        while w < window_size and i >= 0:
            if len({haplotypes[s][i] for s in c_slots}) > 1:
                het_pos.append(i)
                w += 1
            i -= 1
        het_pos = het_pos[::-1] + snps
        # w = number of preceding het positions, can be lower than window_size close to chr start

        # store the allele of every cluster-read for the relevant positions
        rmat = dict()

        # for each position store the supporting reads
        readlist = defaultdict(list)

        # go over all reads and determine which snp positions they are defined on
        for rid in clustering[cid]:
            read = readset[rid]
            rmat[rid] = dict()
            i, j = 0, 0
            while i < len(read) and j < len(het_pos):
                read_pos = index[read[i].position]
                if read_pos == het_pos[j]:
                    rmat[rid][j] = read[i].allele
                    readlist[j].append(rid)
                    i += 1
                    j += 1
                elif read_pos < het_pos[j]:
                    i += 1
                else:
                    j += 1

        # reconstruct cluster phasing
        for i in range(len(snps)):
            pos = snps[i]
            het_idx = i + w
            c_slots = [j for j in range(len(path[pos])) if path[pos][j] == cid]

            configs = []
            # enumerate all permutations of haplotypes at current positions
            for perm in it.permutations(c_slots):
                score = 0
                # for each read: determine number of errors for best fit
                for rid in readlist[het_idx]:
                    if len(rmat[rid]) < 2:
                        continue
                    likelihood = 0.0
                    a_priori = 1 / len(c_slots)
                    for slot in range(len(c_slots)):
                        overlap, errors = 0, 0
                        # skip if read has different allele at current pos than tested config
                        if haplotypes[perm[slot]][pos] != rmat[rid][het_idx]:
                            errors += 1
                        # else, compare to window_size many previous positions
                        for j in range(max(0, het_idx - window_size), het_idx):
                            if j in rmat[rid]:
                                overlap += 1
                                if haplotypes[c_slots[slot]][het_pos[j]] != rmat[rid][j]:
                                    errors += 1
                        likelihood += (
                            a_priori
                            * ((1 - error_rate) ** (overlap - errors))
                            * (error_rate**errors)
                        )
                    score += log(likelihood)
                configs.append((perm, score))

            configs.sort(key=lambda x: -x[1])
            best_perm = configs[0][0]

            # switch alleles at current position, needed as input for next position(s)
            alleles = [haplotypes[best_perm[j]][pos] for j in range(len(best_perm))]
            for j in range(len(c_slots)):
                haplotypes[c_slots[j]][pos] = alleles[j]


def resolve_ambiguous_switches(
    path, haplotypes, clustering, readset, index, error_rate, window_size
):
    """
    Post processing step after the threading. If a haplotype leaves a cluster on a collapsed region,
    we could use read information to find the most likely continuation of the switching haplotypes.
    """
    if len(path) == 0:
        return []

    ploidy = len(path[0])
    events = []
    event_type = ReorderType.NONE

    # maps the haplotype slot of the corrected version to the original (and in reverse)
    current_perm, inverse_perm = list(range(ploidy)), list(range(ploidy))

    for i in range(1, len(path)):
        # find slots affected by cluster changes
        changed_h = {j for j in range(ploidy) if path[i - 1][j] != path[i][inverse_perm[j]]}
        c_group = {path[i - 1][j] for j in changed_h}
        h_group = sorted(j for j in range(ploidy) if path[i - 1][j] in c_group)

        # if only one slot involved -> just write current permutation to haplotype object
        if len(h_group) < 2:
            path[i] = [path[i][j] for j in inverse_perm]
            reord = [haplotypes[inverse_perm[j]][i] for j in range(ploidy)]
            for j in range(ploidy):
                haplotypes[j][i] = reord[j]
            continue

        # compute the optimal permutation of changing haplotypes, according to read information
        for b in [True, False]:
            configs = solve_single_ambiguous_site(
                haplotypes,
                inverse_perm,
                clustering,
                readset,
                index,
                i,
                path,
                h_group,
                c_group,
                b,
                error_rate,
                window_size,
            )
            if len(c_group) == 1:
                event_type = ReorderType.COLLAPSED_PRE if b else ReorderType.COLLAPSED
            else:
                event_type = ReorderType.MULTI
            best_perm = configs[0][0]
            if configs[0][1] > configs[1][1] + 0.5:
                break
        event = ReorderEvent(i, event_type, exp(configs[0][1]), exp(configs[1][1]), h_group)
        events.append(event)

        # apply local best permutation to current global permutation
        current_perm_copy = list(current_perm)
        for j in range(len(h_group)):
            current_perm_copy[inverse_perm[h_group[j]]] = h_group[best_perm[j]]
        current_perm = current_perm_copy
        for j in range(ploidy):
            inverse_perm[current_perm[j]] = j

        # correct current position according to updated permutation
        path[i] = [path[i][j] for j in inverse_perm]
        reord = [haplotypes[inverse_perm[j]][i] for j in range(ploidy)]
        for j in range(ploidy):
            haplotypes[j][i] = reord[j]

    return events


def solve_single_ambiguous_site(
    haplotypes,
    hap_perm,
    clustering,
    readset,
    index,
    pos,
    path,
    h_group,
    c_group,
    exclude_collapsed_cluster,
    error_rate,
    window_size,
):
    # find the last window_size positions (starting from pos - 1),
    # which are heterozyguous among the haplotype group
    het_pos_before, het_pos_after = [], []
    j = pos - 1
    if exclude_collapsed_cluster:
        while len({path[j][h] for h in h_group}) < 2 and j > 0:
            j -= 1

    while len(het_pos_before) < window_size and j >= 0:
        if len({haplotypes[h][j] for h in h_group}) > 1:
            het_pos_before.append(j)
        j -= 1
    het_pos_before = het_pos_before[::-1]
    # same for the next window_size positions (starting from pos)
    j = pos
    while len(het_pos_after) < window_size and j < len(haplotypes[0]):
        if len({haplotypes[hap_perm[h]][j] for h in h_group}) > 1:
            het_pos_after.append(j)
        j += 1
    het_pos = het_pos_before + het_pos_after

    # store the allele of every cluster-read for the relevant positions
    rmat = dict()

    # for each position store the supporting reads
    readlist = []

    # go over all reads and determine which snp positions they are defined on
    if len(het_pos_before) >= 2 and len(het_pos_after) >= 2:
        for cid in c_group:
            for rid in clustering[cid]:
                read = readset[rid]
                if len(read) < 2:
                    continue
                if index[read[0].position] > het_pos_before[-1]:
                    continue
                if index[read[-1].position] < het_pos_after[0]:
                    continue
                rmat[rid] = dict()
                j, k, b, t = 0, 0, 0, 0
                while j < len(read) and k < len(het_pos):
                    read_pos = index[read[j].position]
                    if read_pos == het_pos[k]:
                        rmat[rid][het_pos[k]] = read[j].allele
                        b += 1 if k < len(het_pos_before) else 0
                        t, j, k = t + 1, j + 1, k + 1
                    elif read_pos < het_pos[k]:
                        j += 1
                    else:
                        k += 1

                if b > 0 and t > b:
                    readlist.append(rid)

    # reconstruct cluster phasing
    configs = []
    # enumerate all permutations of haplotypes at current positions
    for perm in it.permutations(range(len(h_group))):
        score = 0.0
        num_factors = 0
        # for each read: determine number of errors for best fit
        for rid in readlist:
            likelihood = 0.0
            for slot in range(len(h_group)):
                overlap = len(rmat[rid])
                errors = 0
                for p in rmat[rid]:
                    if p < het_pos_after[0]:
                        cmp = haplotypes[h_group[perm[slot]]][p]
                    else:
                        cmp = haplotypes[hap_perm[h_group[slot]]][p]
                    if cmp != rmat[rid][p]:
                        errors += 1
                likelihood = max(
                    likelihood, ((1 - error_rate) ** (overlap - errors)) * (error_rate**errors)
                )
            num_factors += overlap
            score += log(likelihood)
        configs.append((perm, score / num_factors if num_factors > 0 else 0))

    configs.sort(key=lambda x: -x[1])
    return configs


def compute_cut_positions(path, block_cut_sensitivity, events):
    """
    Takes a threading as input and computes on which positions a cut should be made according the
    cut sensitivity. The levels mean:

    0 -- No cuts at all, even if regions are not connected by any reads
    1 -- Only cut, when regions are not connected by any reads (already done in advance)
    2 -- Only cut, when regions are not connected by a sufficient number of reads (also done)
    3-5 -- Cut on reordering events, depending on confidence values and types.

    The list of cut positions contains the first position of every block. Therefore, position 0 is
    always in the cut list. The second return value is a list of cut positions for every haplotype
    individually.
    """

    cuts = [0]
    hap_cuts = [[0] for _ in range(len(path[0]) if path else 0)]

    if block_cut_sensitivity >= 3:
        if block_cut_sensitivity >= 5:
            pre_thrshld = 0.9
            mul_thrshld = 1.0
            gen_thrshld = 1.0
        elif block_cut_sensitivity == 4:
            pre_thrshld = 0.7
            mul_thrshld = 0.9
            gen_thrshld = 0.9
        else:
            pre_thrshld = 0.5
            mul_thrshld = 0.8
            gen_thrshld = 0.8

        for e in events:
            cut = False
            if e.event_type is ReorderType.COLLAPSED_PRE and e.confidence <= pre_thrshld:
                cut = True
            elif e.event_type is ReorderType.MULTI and e.confidence <= mul_thrshld:
                cut = True
            elif e.event_type is ReorderType.COLLAPSED and e.confidence <= gen_thrshld:
                cut = True
            if cut:
                cuts.append(e.position)
                for h in e.haplotypes:
                    hap_cuts[h].append(e.position)

    return cuts, hap_cuts



def edit_distance(s: str, t: str, maxdiff: int = ...) -> int: ...
def edit_distance_affine_gap(
    query: str, ref: str, mismatch_cost: int, gap_start: int = ..., gap_extend: int = ...
) -> int: ...



# cython: language_level=3

"""
Edit distance computation.

Copied from https://bitbucket.org/marcelm/sqt/src/af255d54a21815cb9a3e0b279b431a320d4626bd/sqt/_helpers.pyx
"""
from cython.view cimport array as cvarray
import cython
from libc cimport limits

@cython.boundscheck(False)
def edit_distance(s, t, int maxdiff=-1):
	"""
	Return the edit distance between the strings s and t.
	The edit distance is the sum of the numbers of insertions, deletions,
	and mismatches that is minimally necessary to transform one string
	into the other.

	If maxdiff is not -1, then a banded alignment is performed. In that case,
	the true edit distance is returned if and only if it is maxdiff or less.
	Otherwise, a value is returned that is guaranteed to be greater than
	maxdiff, but which is not necessarily the true edit distance.
	"""
	cdef int m = len(s)  # index: i
	cdef int n = len(t)  # index: j
	cdef int e = maxdiff
	cdef int i, j, start, stop, c, prev, smallest
	cdef bint match
	cdef bytes s_bytes, t_bytes
	cdef char* sv
	cdef char* tv


	# Return early if string lengths are too different
	if e != -1 and abs(m - n) > e:
		return abs(m - n)

	s_bytes = s.encode() if isinstance(s, unicode) else s
	t_bytes = t.encode() if isinstance(t, unicode) else t
	sv = s_bytes
	tv = t_bytes

	# Skip identical prefixes
	while m > 0 and n > 0 and sv[0] == tv[0]:
		sv += 1
		tv += 1
		m -= 1
		n -= 1

	# Skip identical suffixes
	while m > 0 and n > 0 and sv[m-1] == tv[n-1]:
		m -= 1
		n -= 1

	cdef int[:] costs = cvarray(shape=(m+1,), itemsize=sizeof(int), format="i")
	if e == -1:
		# Regular (unbanded) global alignment
		with nogil:
			for i in range(m + 1):
				costs[i] = i

			# compute columns of the alignment matrix (using unit costs)
			prev = 0
			for j in range(1, n+1):
				prev = costs[0]
				costs[0] += 1
				for i in range(1, m+1):
					match = sv[i-1] == tv[j-1]
					c = min(
						prev + 1 - match,
						costs[i] + 1,
						costs[i-1] + 1)
					prev = costs[i]
					costs[i] = c
	else:
		# Banded alignment
		with nogil:
			for i in range(m + 1):
				costs[i] = i
			smallest = 0
			for j in range(1, n + 1):
				stop = min(j + e + 1, m + 1)
				if j <= e:
					prev = costs[0]
					costs[0] += 1
					smallest = costs[0]
					start = 1
				else:
					start = j - e
					prev = costs[start - 1]
					smallest = maxdiff + 1
				for i in range(start, stop):
					match = sv[i-1] == tv[j-1]
					c = min(
						prev + 1 - match,
						costs[i] + 1,
						costs[i-1] + 1)
					prev = costs[i]
					costs[i] = c
					smallest = min(smallest, c)
				if smallest > maxdiff:
					break
		if smallest > maxdiff:
			return smallest
	return costs[m]

def f(l, gap_start, gap_ext):
	return gap_start + (l-1) * gap_ext

def edit_distance_affine_gap(query,ref, mismatch_cost, int gap_start=1, int gap_extend=1):
	"""
	Compute edit distance between strings s and t using affine gap costs.
	(gotoh-algorithm)

	gap_start -- cost for starting a gap
	gap_extend -- cost for extending a gap
	mismatch_cost -- list with mismatch costs for each position

	"""

	assert len(query) == len(mismatch_cost)

	cdef int m = len(query)
	cdef int n = len(ref)
	cdef int match_cost = 0
	cdef int i,j
	cdef float prev_a, prev_b, prev_c, m_c, c_a, c_b, c_c
	cdef bytes s_bytes, t_bytes
	cdef char* sv
	cdef char* tv
	cdef len_p = 0

	s_bytes = query.encode() if isinstance(query, unicode) else query
	t_bytes = ref.encode() if isinstance(ref, unicode) else ref
	sv = s_bytes
	tv = t_bytes

	# Skip identical prefixes
	while m > 0 and n > 0 and sv[0] == tv[0]:
		sv += 1
		tv += 1
		m -= 1
		n -= 1
		len_p += 1

	# Skip identical suffixes
	while m > 0 and n > 0 and sv[m-1] == tv[n-1]:
		m -= 1
		n -= 1

	# three DP tables needed
	cdef float[:] a = cvarray(shape=(m+1,), itemsize=sizeof(float), format="f")
	cdef float[:] b = cvarray(shape=(m+1,), itemsize=sizeof(float), format="f")
	cdef float[:] c = cvarray(shape=(m+1,), itemsize=sizeof(float), format="f")

	# initialize
	a[0] = 0
	b[0] = 0
	c[0] = 0

	for i in range(1,m+1):
		a[i] = limits.INT_MAX
		b[i] = f(i, gap_start, gap_extend)
		c[i] = limits.INT_MAX

	# compute tables column wise
	for j in range(1,n+1):
		prev_a = a[0]
		prev_b = b[0]
		prev_c = c[0]

		a[0] = limits.INT_MAX
		b[0] = limits.INT_MAX
		c[0] = f(j, gap_start, gap_extend)

		for i in range(1,m+1):
			m_c = mismatch_cost[i-1 + len_p]
			if sv[i-1] == tv[j-1]:
				m_c = match_cost

			c_a = min(prev_a, prev_b, prev_c) + m_c
			c_b = min(a[i-1] + gap_start, b[i-1] + gap_extend, c[i-1] + gap_start)
			c_c = min(a[i] + gap_start, b[i] + gap_start, c[i] + gap_extend)

			prev_a = a[i]
			prev_b = b[i]
			prev_c = c[i]

			a[i] = c_a
			b[i] = c_b
			c[i] = c_c
	return int(min(a[m], b[m], c[m]))



# cython: language_level=3

from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libcpp.unordered_map cimport unordered_map
from libcpp cimport bool

import math


cdef bool _vector_score_lower(priority_type_ptr first, priority_type_ptr second):
	for i in range(min(first[0].size(), second[0].size())):
		if first[0][i] < second[0][i]:
			return True
		if first[0][i] > second[0][i]:
			return False
	if first[0].size() < second[0].size(): 
		return True
	else:
		return False


cdef priority_type_ptr _pyscore_to_vector(score):
	cdef priority_type_ptr result = new priority_type()
	if isinstance(score,int):
		result[0].push_back(score)
	else:
		try:
			for i in score:
				if not isinstance(i,int):
					raise
				result[0].push_back(i)
		except:
			raise ValueError('Score parameter must be either int, or an iterable object yielding ints')
	return result


cdef int _parent(int index):
	''''returns the index of the parent node in the heap depending on the given index'''
	return (index - 1) // 2


cdef int _left_child(int index):
	'''return the index of the left child in the heap depending on the given index'''
	return (2 * index) + 1


cdef int _right_child(int index):
	'''return the index of the right child in the heap depending on the given index'''
	return (2 * index) + 2


# TODO: in general, see https://www.python.org/dev/peps/pep-0008
cdef class PriorityQueue:
	def __dealloc__(self):
		for i in range(self.heap.size()):
			del self.heap[i].first
	
	def push(self, score, int item):
		'''Add item with given score to the heap.'''
		# Item stored with score in a tupel where the score is always the first position..
		cdef priority_type_ptr c_score = _pyscore_to_vector(score)
		self.c_push(c_score, item)

	cdef void c_push(self, priority_type_ptr score, int item):
		'''Pointer ownership is transferred to priorityqueue.'''
		newindex = self.heap.size()
		cdef queue_entry_type entry
		entry.first = score
		entry.second = item 
		self.heap.push_back(entry)
		self.positions[item] = newindex
		self._sift_up(newindex)

	cdef void _swap(self, int index1, int index2):
		'''swaps the position of the nodes in the priority queue from first index(index1)and second index(index2)'''

		cdef queue_entry_type entry1 = self.heap[index1]
		cdef int pos1 = self.positions[entry1.second]

		cdef queue_entry_type entry2 = self.heap[index2]
		cdef int pos2 = self.positions[entry2.second]

		self.positions[entry1.second] = pos2
		self.positions[entry2.second] = pos1

		self.heap[index1] = entry2
		self.heap[index2] = entry1

	cdef bool _score_lower(self, int index1, int index2):
		cdef queue_entry_type entry1 = self.heap[index1]
		cdef queue_entry_type entry2 = self.heap[index2]
		return _vector_score_lower(entry1.first, entry2.first)

	cdef void _sift_up(self, int index):
		''' recursive method to check if score of item at given index is higher than the score of the parent
		and swaps the nodes till priorityqueue property is restored'''
		#print(str(self), '_sift_up', index)
		cdef int parentindex = _parent(index)
		assert parentindex != index
		if parentindex >= 0:
			if self._score_lower(parentindex, index):
				self._swap(parentindex, index)
				self._sift_up(parentindex)


	cdef void _sift_down(self, int index):
		'''Check if score of item at given index is lower than the score of its children,
		 therefore need to swap position with its children'''
		cdef int rchildindex = _right_child(index)
		cdef int lchildindex = _left_child(index)
		assert rchildindex != index
		assert lchildindex != index
		#if both children are in the heap
		#only need to know if right child exists then automatically also the left child exists
		if rchildindex < self.heap.size():
			if self._score_lower(lchildindex, rchildindex):
				if self._score_lower(index, rchildindex):
					self._swap(rchildindex, index)
					self._sift_down(rchildindex)
			else:
				if self._score_lower(index, lchildindex):
					self._swap(lchildindex, index)
					self._sift_down(lchildindex)
		elif lchildindex < self.heap.size():
			if self._score_lower(index, lchildindex):
				self._swap(lchildindex, index)
				self._sift_down(lchildindex)

	def pop(self):
		'''Removes the item with largest score and returns it as tupel of (score, item).'''
		cdef queue_entry_type entry = self.c_pop()
		cdef priority_type_ptr score = entry.first
		cdef item_type item = entry.second
		if score[0].size() == 1:
			result = (score[0][0], item)
		else:
			result = tuple(score[0]), item
		del score
		return result

	cdef queue_entry_type c_pop(self):
		'''Removes the item with largest score and returns it as tupel of (score, item).'''
		if self.heap.size() == 0:
			raise IndexError('PriorityQueue empty.')
		cdef queue_entry_type last_entry = self.heap[self.heap.size()-1]
		cdef queue_entry_type first_entry = self.heap[0]
		#looks if heap is only one element then no need restore heap property
		if self.heap.size() == 1:
			self.positions.erase(first_entry.second)
			self.heap.pop_back()
		else:
			self.heap[0] = last_entry
			self.heap.pop_back()

			self.positions[last_entry.second]= 0
			self.positions.erase(first_entry.second)

			self._sift_down(0)
		return first_entry

	def change_score(self, item_type item, new_score):
		'''Changes the score of the given item to the new assigned score.'''
		cdef priority_type_ptr c_new_score = _pyscore_to_vector(new_score)
		self.c_change_score(item, c_new_score)

	cdef void c_change_score(self, item_type item, priority_type_ptr c_new_score):
		'''Changes the score of the given item to the new assigned score.
		Ownership of c_new_score goes to priorityqueue.'''
		position = self.positions[item]

		c_old_score = self.heap[position].first
		self.heap[position].first = c_new_score

		# Differentiate between increasing and decreasing score
		if _vector_score_lower(c_old_score, c_new_score):
			self._sift_up(position)
		else:
			self._sift_down(position)

		del c_old_score

	def get_score_by_item(self, item):
		'''returns actual score of the given item or None if item is not in the heap '''
		cdef priority_type_ptr score = self.c_get_score_by_item(item)
		if score == NULL:
			return None
		elif score[0].size() == 1:
			return score[0][0]
		else:
			return tuple(score[0])

	cdef priority_type_ptr c_get_score_by_item(self, item_type item):
		'''Returns score of the given item or NULL if item is not in the heap. 
		Pointer ownership stays with priorityqueue.'''
		cdef unordered_map[item_type,int].iterator it = self.positions.find(item)
		if it == self.positions.end():
			return NULL
		cdef queue_entry_type entry = self.heap[self.positions[item]]
		return entry.first
		cdef priority_type_ptr score = entry.first
		if score[0].size() == 1:
			return score[0][0]
		else:
			return tuple(score[0])

	def __len__(self):
		'''Length of  Priority Queue is equal to the length of the stored heap'''
		return self.heap.size()

	cdef int size(self):
		return self.heap.size()

	def is_empty(self):
		'''Return if actual Priority Queue is Empty'''
		return self.heap.size() == 0

	cdef bool c_is_empty(self):
		return self.heap.size() == 0
